"""
Neuraxle's AutoML Classes
====================================
Classes used to build any Automatic Machine Learning pipelines.
Hyperparameter selection strategies are used to optimize the hyperparameters of given pipelines.

..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import gc
from copy import copy
from typing import ContextManager, Iterator, List, Optional, Tuple

from neuraxle.base import (CX, BaseService, BaseStep, BaseStepT, ExecutionContext, ExecutionMode, ForceHandleMixin,
                           TruncableService, _HasChildrenMixin)
from neuraxle.data_container import ARG_X_INPUTTED, IDT
from neuraxle.data_container import DataContainer as DACT
from neuraxle.data_container import TrainDACT
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.callbacks import (ARG_Y_EXPECTED, ARG_Y_PREDICTD, BaseCallback, CallbackList, MetricCallback,
                                        ScoringCallback)
from neuraxle.metaopt.context import AutoMLContext
from neuraxle.metaopt.data.aggregates import Client, Project, Root, Round, Trial, TrialSplit
from neuraxle.metaopt.data.reporting import RoundReport
from neuraxle.metaopt.data.vanilla import DEFAULT_CLIENT, DEFAULT_PROJECT, RoundDataclass, ScopedLocation
from neuraxle.metaopt.optimizer import BaseHyperparameterOptimizer, GridExplorationSampler, RandomSearchSampler
from neuraxle.metaopt.repositories.repo import HyperparamsRepository
from neuraxle.metaopt.validation import BaseValidationSplitter, ValidationSplitter


class Trainer(BaseService):
    """
    Class used to train a pipeline using various data splits and callbacks for evaluation purposes.
    It loops on splits that the splitter yields, and on epochs as well, to train and validate the pipeline
    with the given metrics and other callbacks.

    If the predicted expected output of a pipeline's prediction dact is not empty, then it will be used
    in the metrics instead of using the fed dacts' expected output. This is to allow for the use of
    autoregressive models, where the expected output is not known at the time of sending it to the model at train-time, but is known for the least at validation time as per the validation splitter.
    """
    # TODO: add this `with_val_set` method that would change splitter to
    #       PresetValidationSetSplitter(self, val) and override..?

    def __init__(
            self,
            validation_splitter: 'BaseValidationSplitter',
            callbacks: List[BaseCallback] = None,
            n_epochs: int = 1,
    ):
        BaseService.__init__(self)
        self.validation_splitter: BaseValidationSplitter = validation_splitter
        self.callbacks: CallbackList = CallbackList(callbacks) if callbacks is not None else []
        self.n_epochs = n_epochs

    def train(
        self,
        pipeline: BaseStep,
        dact: DACT,
        trial_scope: Trial,
        return_trained_pipelines: bool = False
    ) -> Optional[List[BaseStep]]:
        """
        Train pipeline using the validation splitter.
        Track training, and validation metrics for each epoch.
        Note: the present method is just a shortcut to using the `execute_trial` method with less boilerplate code needed. Refer to `execute_trial` for full flexibility
        """
        trained_pipelines: List[BaseStep] = []

        splits: List[Tuple[DACT, DACT]] = self.validation_splitter.split_dact(
            dact, context=trial_scope.context)

        for train_dact, val_dact in splits:
            with trial_scope.new_validation_split() as trial_split_scope:
                p = self.train_split(pipeline, train_dact, val_dact, trial_split_scope)
                if return_trained_pipelines:
                    trained_pipelines.append(p)
                else:
                    del p
                    gc.collect()

        return trained_pipelines

    def train_split(
        self,
        pipeline: BaseStep,
        train_dact: DACT,
        val_dact: Optional[DACT],
        trial_split_scope: TrialSplit
    ) -> BaseStep:
        """
        Train a pipeline split. You probably want to use `self.train` instead, to use the validation splitter.
        If validation DACT is None, the evaluation metrics will not save validation results.

        It is to be noted that here, if the data container, after a prediction at train and validation time,
        has an empty expected output (of .expected_outputs of length 0 or that is None), then the
        trainer will pick the expected output of the pre-predicted data container that was fed as an input.
        """
        trial_split_scope: TrialSplit = trial_split_scope.with_n_epochs(self.n_epochs)
        p: BaseStep = pipeline._copy(trial_split_scope.context, deep=True)
        p.set_hyperparams(trial_split_scope.get_hyperparams())
        context: AutoMLContext = trial_split_scope.context

        for _ in range(self.n_epochs):
            e = trial_split_scope.next_epoch()

            # Fit train
            p = p.set_train(True)
            context = context.train()
            p = p.handle_fit(
                train_dact.copy(),
                context
            )

            # Predict train & val
            p = p.set_train(False)
            context = context.validation()
            eval_dact_train = p.handle_predict(
                train_dact.without_eo(),
                context
            )
            eval_dact_train: DACT[IDT, ARG_Y_PREDICTD, ARG_Y_EXPECTED] = eval_dact_train
            _has_empty_eo = eval_dact_train.expected_outputs is None or (hasattr(
                eval_dact_train.expected_outputs, "__len__") and len(eval_dact_train.expected_outputs) == 0)
            if _has_empty_eo or self.validation_splitter.force_fixed_metric_expected_outputs is True:
                eval_dact_train = eval_dact_train.with_eo(train_dact.expected_outputs)

            if val_dact is not None:
                context = context.validation()
                eval_dact_valid = p.handle_predict(
                    val_dact.without_eo(),
                    context
                )
                eval_dact_valid: DACT[IDT, ARG_Y_PREDICTD, ARG_Y_EXPECTED] = eval_dact_valid
                _has_empty_eo = eval_dact_valid.expected_outputs is None or (hasattr(
                    eval_dact_valid.expected_outputs, "__len__") and len(eval_dact_valid.expected_outputs) == 0)
                if _has_empty_eo or self.validation_splitter.force_fixed_metric_expected_outputs is True:
                    eval_dact_valid = eval_dact_valid.with_eo(val_dact.expected_outputs)
            else:
                eval_dact_valid = None

            # Log metrics / evaluate
            if self.callbacks.call(
                trial_split_scope,
                eval_dact_train,
                eval_dact_valid,
                e == self.n_epochs
            ):
                break  # Saves stats using the '__exit__' method of managed scoped aggregates.

        return p

    def refit(
        self,
        pipeline: BaseStep,
        dact: DACT,
        trial_scope: Trial,
    ) -> BaseStep:
        """
        Refit the pipeline on the whole dataset (without any validation technique).

        :return: fitted pipeline
        """

        with trial_scope.retrain_split() as trial_split_scope:
            trial_split_scope: TrialSplit = trial_split_scope

            return self.train_split(
                pipeline,
                dact,
                None,
                trial_split_scope
            )


class BaseControllerLoop(TruncableService):

    def __init__(
        self,
        trainer: Trainer,
        n_trials: int,
        hp_optimizer: BaseHyperparameterOptimizer = None,
        continue_loop_on_error: bool = True
    ):
        if hp_optimizer is None:
            hp_optimizer = RandomSearchSampler()

        TruncableService.__init__(self, {
            Trainer: trainer,
            BaseHyperparameterOptimizer: hp_optimizer,
        })
        self.n_trials = n_trials
        self.continue_loop_on_error: bool = continue_loop_on_error

    @property
    def trainer(self) -> Trainer:
        return self.get_service(Trainer)

    def run(self, pipeline: BaseStep, dact: DACT, round_scope: Round):
        """
        Run the controller loop.

        :param context: execution context
        :return: the ID of the round that was executed (either created or continued from previous optimization).
        """

        # thread_safe_lock, ..., context = context.thread_safe() ??

        hp_optimizer: BaseHyperparameterOptimizer = self[BaseHyperparameterOptimizer]
        hp_space: HyperparameterSpace = pipeline.get_hyperparams_space()

        round_scope: Round = round_scope.with_optimizer(hp_optimizer, hp_space)

        for managed_trial_scope in self.loop(round_scope):
            managed_trial_scope: Trial = managed_trial_scope  # typing helps
            # trial_scope.context.restore_lock(thread_safe_lock) ??
            self.trainer.train(pipeline, dact, managed_trial_scope)

    def loop(self, round_scope: Round) -> Iterator[Trial]:
        """
        Loop over all trials.

        :param dact: data container that is not yet splitted
        :param context: execution context
        :return:
        """
        for _ in range(self.n_trials):
            with self.next_trial(round_scope) as managed_trial_scope:
                managed_trial_scope: Trial = managed_trial_scope  # typing helps

                yield managed_trial_scope

    def next_trial(self, round_scope: Round) -> ContextManager[Trial]:
        """
        Get the next trial to be executed.

        :param round_scope: round scope
        :return: the next trial to be executed.
        """
        return round_scope.new_rvs_trial(self.continue_loop_on_error)  # TODO: parallelization with this method?

    def refit_best_trial(self, pipeline: BaseStep, dact: DACT, round_scope: Round) -> BaseStep:
        """
        Refit the pipeline on the whole dataset (without any validation technique).
        """
        with round_scope.refitting_best_trial() as managed_trial_scope:

            refitted: BaseStep = self.trainer.refit(
                pipeline, dact, managed_trial_scope)

        return refitted

    def for_refit_only(self) -> 'BaseControllerLoop':
        """
        Create a controller loop configured with zero iterations
        so as to only make the "refit_best_trial" possible.
        """
        self_copy = copy(self)
        self_copy.n_trials = 0
        return self_copy


class DefaultLoop(BaseControllerLoop):

    def __init__(
        self,
        trainer: Trainer,
        n_trials: int,
        hp_optimizer: BaseHyperparameterOptimizer = None,
        continue_loop_on_error: bool = False,
        n_jobs: int = 1,
    ):
        super().__init__(
            trainer,
            n_trials,
            hp_optimizer=hp_optimizer,
            continue_loop_on_error=continue_loop_on_error
        )
        self.n_jobs = n_jobs


class ControlledAutoML(ForceHandleMixin, _HasChildrenMixin[BaseStepT], BaseStep):
    """
    A step to execute Automated Machine Learning (AutoML) algorithms. This step will
    automatically split the data into train and validation splits, and execute an
    hyperparameter optimization on the splits to find the best hyperparameters.

    The :class:`BaseControllerLoop` is useful to possibly split the execution into multiple
    threads, or even multiple machines to decide how to execute the loop.

    The :class:`Trainer` is responsible for training the pipeline on the train and validation
    splits, as per the data split provided by the splitter, and the predicted data containers.

    It is to be noted that if the data container, after a prediction at train and validation time,
    has an empty expected output (of .expected_outputs of length 0 or that is None), then the
    trainer will pick the expected output of the pre-predicted data container that was fed as an input.

    The step with the chosen best hyperparameters will be optionnally refitted to the full
    unsplitted data (pre-split data) if desired, and will be useable using :func:`refit_best_trial`.
    """

    def __init__(
            self,
            pipeline: BaseStepT,
            loop: BaseControllerLoop,
            main_metric_name: str,
            repo: HyperparamsRepository = None,
            start_new_round: bool = True,
            refit_best_trial: bool = True,
            project_name: str = DEFAULT_PROJECT,
            client_name: str = DEFAULT_CLIENT,
    ):
        """
        .. note::
            Usage of a multiprocess-safe hyperparams repository is recommended,
            although it is, most of the time, not necessary.
            Context instances are not shared between trial but copied.
            So is the AutoML loop and the DACTs.

        :param pipeline: The pipeline, or BaseStep, which will be use by the AutoMLloop
        :param loop: The loop, or BaseControllerLoop, which will be used by the AutoML loop
        :param flow: The flow, or Flow, which will be used by the AutoML loop
        :param refit_best_trial: A boolean indicating whether to perform, after a fit call, a refit on the best trial.
        """
        BaseStep.__init__(self)
        _HasChildrenMixin.__init__(self)
        ForceHandleMixin.__init__(self)

        self.pipeline: BaseStepT = pipeline
        self.loop: BaseControllerLoop = loop
        self.repo: HyperparamsRepository = repo

        self.main_metric_name: str = main_metric_name
        self.start_new_round: bool = start_new_round
        self.refit_best_trial: bool = refit_best_trial
        self.project_name: str = project_name
        self.client_name: str = client_name
        self._round_number: Optional[int] = None

        self.has_model_been_retrained: bool = False

    def get_children(self) -> List[BaseStep]:
        return [self.pipeline]

    @property
    def wrapped(self) -> BaseStep:
        return self.pipeline

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseStep', DACT]:
        if not self.refit_best_trial:
            raise ValueError(
                "self.refit_best_trial must be True in this AutoML class to do the transform in 'fit_transform'.")

        self = self._fit_data_container(data_container, context)
        data_container = self._transform_data_container(data_container, context)
        return self, data_container

    def to_force_refit_best_trial(self) -> 'ControlledAutoML':
        self_copy = copy(self)
        self_copy.refit_best_trial = True
        self_copy.has_model_been_retrained = False
        self_copy.start_new_round = False
        self_copy.loop = self_copy.loop.for_refit_only()
        return self_copy

    def _fit_data_container(self, data_container: DACT, context: CX) -> 'BaseStep':
        """
        Run Auto ML Loop.
        Find the best hyperparams using the hyperparameter optmizer.
        Evaluate the pipeline on each trial using a validation technique.

        :param data_container: data container to fit
        :param context: execution context

        :return: self
        """
        automl_context: AutoMLContext = self.get_automl_context(context, with_loc=False)
        root: Root = Root.from_context(automl_context, is_deep=False)

        with root.get_project(self.project_name) as ps:
            ps: Project = ps
            with ps.get_client(self.client_name) as cs:
                cs: Client = cs
                with cs.optim_round(self.start_new_round, self.main_metric_name) as rs:
                    rs: Round = rs
                    self._round_number = rs.round_number

                    self.loop.run(self.pipeline, data_container, rs)

                    if self.refit_best_trial:
                        self.pipeline = self.loop.refit_best_trial(self.pipeline, data_container, rs)
                        self.has_model_been_retrained = True

        return self

    def get_automl_context(self, context: ExecutionContext, with_loc=True) -> AutoMLContext:
        cx = AutoMLContext.from_context(context, repo=self.repo)
        if self.repo is None:
            self.repo = cx.repo
        if with_loc and (self._round_number is not None or not self.start_new_round):
            loc = ScopedLocation(self.project_name, self.client_name, self.round_number)
            cx = cx.with_loc(loc)
        return cx

    def _encapsulate_data(
        self, data_inputs: ARG_X_INPUTTED, expected_outputs: ARG_Y_EXPECTED, execution_mode: ExecutionMode
    ) -> Tuple[CX, TrainDACT]:
        """
        This method is overriden from :class:`ForceHandleMixin` to encapsulate
        the repository in a AutoMLContext instead of in a regular CX in case the
        handler methods were not called but the repository was passed at construction.
        """
        data_container = TrainDACT(data_inputs=data_inputs, expected_outputs=expected_outputs)
        context = CX(execution_mode=execution_mode)
        context = self.get_automl_context(context)
        return context, data_container

    def get_best_model(self) -> BaseStep:
        """
        Get the best model if it has been refit, otherwise raises an assertion error.
        """
        self._assert(
            self.has_model_been_retrained,
            "The model has not been retrained, so it cannot be used to get the best model."
        )
        return self.pipeline

    @property
    def round_number(self) -> Optional[int]:
        if self._round_number is None:
            if self.start_new_round:
                raise ValueError("AutoML loop has not been run yet, cannot ask for report.")
            else:
                return len(self.repo.load(ScopedLocation(self.project_name, self.client_name))) - 1
        return self._round_number

    @property
    def report(self) -> RoundReport:
        dc: RoundDataclass = self.repo.load(ScopedLocation(
            self.project_name, self.client_name, self.round_number), deep=True)
        return RoundReport(dc)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        if not self.has_model_been_retrained:
            raise ValueError(
                'self.refit_best_trial must be True in AutoML class '
                'to transform, and the AutoMl should have been fitted.')

        return self.pipeline.handle_transform(data_container, context)


class AutoML(ControlledAutoML):
    """
    This class provides a nice interface to easily use the
    ControlledAutoML class and the metaopt module in general.

    It is a wrapper around the :class:`ControlledAutoML` class, which is a wrapper around
    the AutoML loop contained in the :class:`BaseControllerLoop` class
    and the :class:`Trainer` class all contained in this module.

    :param pipeline: pipeline to copy and use for training
    :param validation_splitter: validation splitter to use
    :param refit_best_trial: whether to refit the best model on the whole dataset after the optimization
    :param scoring_callback: main callback to use for scoring, that is deprecated
    :param hyperparams_optimizer: hyperparams optimizer to use
    :param hyperparams_repository: hyperparams repository to use
    :param n_trials: number of trials to run
    :param epochs: number of epochs to train the model for each val split
    :param callbacks: callbacks to use for training - there can be aditionnal metrics there
    :param n_jobs: number of jobs to use for parallelization, defaults is None for no parallelization
    :param continue_loop_on_error: whether to continue the main optimization loop on error or not
    :return: AutoML object ready to use with fit and transform.
    """

    def __init__(
        self,
        pipeline: BaseStep,
        validation_splitter: 'BaseValidationSplitter' = None,
        hyperparams_optimizer: BaseHyperparameterOptimizer = None,
        scoring_callback: ScoringCallback = None,
        callbacks: List[BaseCallback] = None,
        hyperparams_repository: HyperparamsRepository = None,
        n_trials: int = None,
        refit_best_trial: bool = True,
        start_new_round=True,
        epochs: int = 1,
        n_jobs=1,
        continue_loop_on_error=True
    ):
        # parse or guess args:
        validation_splitter = validation_splitter or ValidationSplitter(0.2)
        hyperparams_optimizer = hyperparams_optimizer or GridExplorationSampler()

        callbacks = list(callbacks) if callbacks is not None else []
        if scoring_callback is not None:
            callbacks = [scoring_callback] + callbacks
        if len(callbacks) == 0:
            raise ValueError("At least one callback must be provided.")
        if not isinstance(callbacks[0], MetricCallback):
            raise ValueError("The first callback is the scoring callback and it must be a MetricCallback.")

        if n_trials is None or n_trials < 1:
            n_trials: int = GridExplorationSampler.estimate_ideal_n_trials(pipeline.get_hyperparams_space())

        # init subservices:
        trainer = Trainer(
            callbacks=callbacks,
            validation_splitter=validation_splitter,
            n_epochs=epochs,
        )
        controller_loop = DefaultLoop(
            trainer=trainer,
            n_trials=n_trials,
            n_jobs=n_jobs,
            hp_optimizer=hyperparams_optimizer,
            continue_loop_on_error=continue_loop_on_error,
        )

        # init base class:
        ControlledAutoML.__init__(
            self,
            pipeline=pipeline,
            loop=controller_loop,
            repo=hyperparams_repository,
            main_metric_name=callbacks[0].name,
            start_new_round=start_new_round,
            refit_best_trial=refit_best_trial,
        )



"""
Neuraxle's Deprecation Warnings
====================================
Code evolves through time. When updating Neuraxle, you may find
that some old arguments you were using or some classes you were
using changed. Warnings will be printed using the methods here.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import warnings


SILENCE_DEPRECATION_WARNING = False


def silence_all_deprecation_warnings():
    """
    Turn off all of the neuraxle.logging.warnings for deprecations.
    """
    global SILENCE_DEPRECATION_WARNING
    SILENCE_DEPRECATION_WARNING = True


def warn_deprecated_class(self, replacement_class: type = None, as_per_version: str = None):
    global SILENCE_DEPRECATION_WARNING
    if not SILENCE_DEPRECATION_WARNING and replacement_class is not None:
        warnings.warn(
            _deprecated_class_msg(replacement_class, as_per_version) + _deact_msg_instructions()
        )
    return self


class RaiseDeprecatedClass:
    """
    Use this class and call it's __init__ method to raise a
    DeprecationWarning and point to the good class to use.
    """

    def __init__(self, replacement_class: type = None, since_version: str = None) -> None:
        raise_deprecated_class(replacement_class, since_version)


def raise_deprecated_class(replacement_class: type = None, since_version: str = None):
    raise DeprecationWarning(_deprecated_class_msg(replacement_class, since_version))


def _deprecated_class_msg(self, replacement_class: type = None, since_version: str = None) -> str:
    return (
        f"The class `{self.__class__.__name__}` is deprecated"
        f" since version `neuraxle>={since_version}`." if since_version is not None else "."
        f" Please consider using the class `{replacement_class.__name__}` instead: visit https://www.neuraxle.org/stable/search.html?q={replacement_class.__name__} for more information" if (
            hasattr(replacement_class, "__name__") and replacement_class.__name__ is not None) else " Visit https://www.neuraxle.org/stable/api.html for more information."
    )


def warn_deprecated_arg(self, arg_name, default_value, value, replacement_argument_name, replacement_class: type = None):
    global SILENCE_DEPRECATION_WARNING
    if not SILENCE_DEPRECATION_WARNING and default_value != value:
        if isinstance(replacement_class, type):
            replacement_class = replacement_class.__name__
        warnings.warn(
            f"Argument `{arg_name}={value}` for class `{self.__class__.__name__}` is deprecated. "
            f"Please consider using `{replacement_argument_name}` "
            f"or the class `{replacement_class}` " if replacement_class is not None else ""
            f"instead. "
            f"{_deact_msg_instructions()}"
        )
    return self


def _deact_msg_instructions() -> str:
    return (
        " If you want to disable these warnings,"
        " call `neuraxle.logging.warnings.silence_all_deprecation_warnings()`."
    )



"""
Neuraxle's SQLAlchemy Hyperparameter Repository Classes
========================================================
Data objects and related repositories used by AutoML, SQL version.

..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
import os
import typing
from datetime import datetime
from typing import Any, Dict, List, Optional, OrderedDict, Type

from neuraxle.base import TrialStatus
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.logging.logging import NeuraxleLogger
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, DEFAULT_METRIC_NAME,
                                           DEFAULT_PROJECT, BaseDataclass,
                                           BaseTrialDataclassMixin,
                                           ClientDataclass,
                                           MetricResultsDataclass,
                                           ProjectDataclass, RootDataclass,
                                           RoundDataclass, ScopedLocation,
                                           SubDataclassT, TrialDataclass,
                                           TrialSplitDataclass,
                                           dataclass_2_id_attr)
from neuraxle.metaopt.repositories.repo import HyperparamsRepository
from sqlalchemy import (JSON, TEXT, Boolean, Column, DateTime, Float,
                        ForeignKey, Integer, String, and_, create_engine)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import (backref, declarative_mixin, joinedload,
                            relationship, sessionmaker)
from sqlalchemy.orm.session import Session
from sqlalchemy.schema import Sequence
from sqlalchemy.sql import asc, desc, func

Base = declarative_base()


class ScopedLocationTreeNode(Base):
    __tablename__ = "scopedlocationtreenode"

    id = Column(Integer, Sequence('id_seq', metadata=Base.metadata), primary_key=True, index=True)
    # TODO: id to be a foreing key in the node.
    parent_id = Column(Integer, ForeignKey(id))

    project_name = Column(String, nullable=True, index=True, default=None)  # DEFAULT_PROJECT
    client_name = Column(String, nullable=True, index=True, default=None)  # DEFAULT_CLIENT
    round_number = Column(Integer, nullable=True, index=True, default=None)
    trial_number = Column(Integer, nullable=True, index=True, default=None)
    split_number = Column(Integer, nullable=True, index=True, default=None)
    metric_name = Column(String, nullable=True, index=True, default=None)

    dataclass_node = relationship("DataClassNode", back_populates="tree", uselist=False)

    # Children:
    subdataclasses = relationship(
        "ScopedLocationTreeNode",
        backref=backref("parent", remote_side=id),
        # collection_class=attribute_mapped_collection("dataclass_node.id_attr_value"),
    )

    def __init__(
        self, dataclass_node: 'DataClassNode', parent: 'ScopedLocationTreeNode' = None
    ):

        if parent is not None:
            self._set_location_from(dataclass_node, parent)

        self.dataclass_node = dataclass_node

        self.parent = parent
        # TODO: parent push self???
        # if parent is not None:
        #    self.parent.subdataclasses.set(self)

    def _set_location_from(self, dataclass_node: 'DataClassNode', parent=None):
        # Set the location of the dataclass node partially from the parent:
        self.project_name = parent.project_name
        self.client_name = parent.client_name
        self.round_number = parent.round_number
        self.trial_number = parent.trial_number
        self.split_number = parent.split_number
        self.metric_name = parent.metric_name

        # Then override the location of the dataclass node:
        dc_klass: Type[BaseDataclass] = sqlalchemy_node_2_dataclass[dataclass_node.__class__]
        id_name: str = dataclass_2_id_attr[dc_klass]
        setattr(self, id_name, str(getattr(dataclass_node, id_name)))

    @property
    def loc(self):
        return ScopedLocation(
            project_name=self.project_name,
            client_name=self.client_name,
            round_number=self.round_number,
            trial_number=self.trial_number,
            split_number=self.split_number,
            metric_name=self.metric_name,
        )

    @property
    def id_attr_value(self):
        return self.dataclass_node.id_attr_value

    def to_dataclass(self, deep=False):
        dc: SubDataclassT = self.dataclass_node.to_empty_dataclass()

        if dc.is_terminal_leaf():
            pass
        elif deep:
            items = [
                (v.id_attr_value, v.to_dataclass(deep=deep)) for v in self.subdataclasses
            ]
            dc.set_sublocation_items(items)
        else:
            dc.set_sublocation_keys([i.id_attr_value for i in self.subdataclasses])

        return dc

    # def update_dataclass(self, _dataclass: SubDataclassT, deep: bool):
    #     if not deep:
    #         _dataclass = _dataclass.empty()
    #     self.dataclass_node.update_dataclass(_dataclass)

    # def add_subdataclass(self, _dataclass: SubDataclassT, deep: bool):
    #     if not deep:
    #         _dataclass = _dataclass.empty()
    #     DataClassNode.add_dataclass(_dataclass=_dataclass, parent=self)

    @staticmethod
    def query(session: Session, scope: ScopedLocation, deep=False) -> Optional['ScopedLocationTreeNode']:
        query = session.query(ScopedLocationTreeNode).filter(
            and_(*[
                getattr(ScopedLocationTreeNode, attr) == str(getattr(scope, attr))
                if getattr(scope, attr) is not None else getattr(ScopedLocationTreeNode, attr) == None
                for attr in ScopedLocation.__dataclass_fields__
            ])
        )

        if deep:
            n_levels_deeper_to_fetch = 6 - len(scope)
            if n_levels_deeper_to_fetch > 0:
                jl = joinedload(ScopedLocationTreeNode.subdataclasses)
                for _ in range(n_levels_deeper_to_fetch - 1):
                    jl = jl.joinedload(ScopedLocationTreeNode.subdataclasses)
                query = query.options(jl)

        query = query.all()

        # the return value is optional:
        if len(query) == 0:
            return None  # TODO: create directly instead?
        elif len(query) == 1:
            return query[0]
        else:
            raise ValueError(f"More than one ScopedLocationTreeNode found for the given scope {scope}: {query}")

    def __str__(self):
        return f"<{self.__class__.__name__}({self.loc.as_list()})>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.loc.as_list()}, ...)>"


class DataClassTreeNodeUpdater:

    def __init__(
        self, session: Session, scope: ScopedLocation, deep: bool = False,
    ):
        self.session = session
        self.scope = scope
        self.deep = deep

    def add_or_update_dataclass(
        self, _dataclass: SubDataclassT, parent: Optional[ScopedLocationTreeNode]
    ):
        """
        Add or update a dataclass in the database.

        Possible scenarios:
        - dataclass is not in the database:
            - add it to the database. Recursively:
                - add all subdataclasses
        - dataclass is in the database:
            - update it in the database. Recursively, for all subdataclasses wheter found or not:
                - add all subdataclasses
                - update all subdataclasses

        This means we have the following methods chaining into each other:
        - add or update dataclass
        - update dataclass
        - add dataclass
        """
        node: ScopedLocationTreeNode = ScopedLocationTreeNode.query(self.session, self.scope, self.deep)
        if node is not None:
            # The dataclass is already in the database. Update it:
            self.update_dataclass(_dataclass, node)
        else:
            # Add the dataclass to the database:
            self.add_dataclass(_dataclass, parent)

    def update_dataclass(
        self, _dataclass: SubDataclassT, node: ScopedLocationTreeNode
    ):
        self.scope = self.scope.with_dc(_dataclass)

        node.dataclass_node._update_attrs_from_dataclass(_dataclass)
        self.session.commit()  # TODO: REMOVE THAT DEBUGGING.

        if self.deep and not _dataclass.is_terminal_leaf():
            _ids_stored: List[str] = [v.id_attr_value for v in node.subdataclasses]
            for sub_dc_key, sub_dc in _dataclass.get_sublocation_items():
                sub_dc_key = str(sub_dc_key)
                if sub_dc_key in _ids_stored:
                    # TODO: what if parent wasn't deep? That is needed to access the subdataclasses.
                    self.update_dataclass(sub_dc, node.subdataclasses[_ids_stored.index(sub_dc_key)])
                else:
                    self.add_dataclass(sub_dc, node)

    def add_dataclass(
        self, _dataclass: SubDataclassT, parent: Optional[ScopedLocationTreeNode]
    ):
        self.scope = self.scope.with_dc(_dataclass)

        # Create a new dataclass node:
        _dc_klass: Type[DataClassNode] = dataclass_2_sqlalchemy_node[_dataclass.__class__]
        dc_node = _dc_klass(_dataclass=_dataclass)

        # Add the dataclass node to the tree's scope:
        parent = parent or ScopedLocationTreeNode.query(self.session, self.scope.popped(), deep=False)
        tree_node = ScopedLocationTreeNode(
            dataclass_node=dc_node,
            parent=parent  # Linked parent makes the new node added to the session automatically.
        )
        self.session.commit()  # TODO: REMOVE THAT DEBUGGING.

        if self.deep and not _dataclass.is_terminal_leaf():
            for sub_dc in _dataclass.get_sublocation_values():
                self.add_dataclass(sub_dc, tree_node)


class DataClassNode(Base):
    __tablename__ = 'dataclassnode'
    # id = Column(String, primary_key=True, nullable=True)
    # id = Column(String, ForeignKey('scopedlocationtreenode.id'), primary_key=True)
    id_attr_value = Column(String, nullable=True)

    type = Column(String(50))

    # tree_id = Column(ForeignKey('scopedlocationtreenode.id'))
    tree_id = Column(String, ForeignKey('scopedlocationtreenode.id'), primary_key=True)
    tree = relationship("ScopedLocationTreeNode", back_populates="dataclass_node")

    logs = relationship("AutoMLLog", back_populates="dataclass_node")

    __mapper_args__ = {
        'polymorphic_identity': 'rootdataclass',
        'polymorphic_on': type
    }

    def __init__(self, _dataclass: BaseDataclass = None):
        if _dataclass is None:
            _dataclass = sqlalchemy_node_2_dataclass[self.__class__]()
        self._update_attrs_from_dataclass(_dataclass)
        self.id_attr_value = str(_dataclass.get_id())

    def _update_attrs_from_dataclass(self, _dataclass: BaseDataclass):
        pass

    def to_empty_dataclass(self) -> RootDataclass:
        return RootDataclass()


class ProjectNode(DataClassNode):
    __tablename__ = 'project'
    id = Column(String, ForeignKey('dataclassnode.tree_id'), primary_key=True)
    project_name = Column(String, nullable=False, default=DEFAULT_PROJECT)

    __mapper_args__ = {
        'polymorphic_identity': 'projectdataclass',
    }

    def _update_attrs_from_dataclass(self, _dataclass: ProjectDataclass):
        self.project_name = _dataclass.project_name

    def to_empty_dataclass(self) -> ProjectDataclass:
        return ProjectDataclass(project_name=self.project_name)


class ClientNode(DataClassNode):
    __tablename__ = 'client'
    id = Column(String, ForeignKey('dataclassnode.tree_id'), primary_key=True)
    client_name = Column(String, nullable=False, default=DEFAULT_CLIENT)

    __mapper_args__ = {
        'polymorphic_identity': 'clientdataclass',
    }

    def _update_attrs_from_dataclass(self, _dataclass: ClientDataclass):
        self.client_name = _dataclass.client_name

    def to_empty_dataclass(self) -> ClientDataclass:
        return ClientDataclass(client_name=self.client_name)


class RoundNode(DataClassNode):
    __tablename__ = 'round'
    # TODO: is it string or int? for all DC Nodes ids.
    id = Column(String, ForeignKey('dataclassnode.tree_id'), primary_key=True)
    round_number = Column(Integer, nullable=False, default=0)
    main_metric_name = Column(String, nullable=True, default=None)

    __mapper_args__ = {
        'polymorphic_identity': 'rounddataclass',
    }

    def _update_attrs_from_dataclass(self, _dataclass: RoundDataclass):
        self.round_number = _dataclass.round_number
        self.main_metric_name = _dataclass.main_metric_name

    def to_empty_dataclass(self) -> RoundDataclass:
        return RoundDataclass(round_number=self.round_number, main_metric_name=self.main_metric_name)


@declarative_mixin
class BaseTrialNodeMixin:
    hyperparams = Column(JSON, nullable=False, default=dict)
    status = Column(String, nullable=False, default=TrialStatus.PLANNED.value)
    created_time = Column(DateTime, nullable=False, default=func.now())
    start_time = Column(DateTime, nullable=True, default=func.now())
    end_time = Column(DateTime, nullable=True)

    def _update_attrs_from_dataclass(self, _dataclass: BaseTrialDataclassMixin):
        self.hyperparams = dict(_dataclass.hyperparams.to_flat_dict())
        self.status = _dataclass.status.value
        self.created_time = _dataclass.created_time
        self.start_time = _dataclass.start_time
        self.end_time = _dataclass.end_time

    def _attrs_for_to_dataclass(self) -> Dict[str, Any]:
        return {
            'hyperparams': HyperparameterSamples(self.hyperparams),
            'status': TrialStatus(self.status),
            'created_time': self.created_time,
            'start_time': self.start_time,
            'end_time': self.end_time,
        }


class TrialNode(BaseTrialNodeMixin, DataClassNode):
    __tablename__ = 'trial'
    id = Column(String, ForeignKey('dataclassnode.tree_id'), primary_key=True)
    trial_number = Column(Integer, nullable=False, default=0)

    __mapper_args__ = {
        'polymorphic_identity': 'trialdataclass',
    }

    def _update_attrs_from_dataclass(self, _dataclass: TrialDataclass):
        BaseTrialNodeMixin._update_attrs_from_dataclass(self, _dataclass=_dataclass)
        self.trial_number = _dataclass.trial_number

    def to_empty_dataclass(self) -> TrialDataclass:
        return TrialDataclass(
            trial_number=self.trial_number, retrained_split=None,
            **self._attrs_for_to_dataclass()
        )


class TrialSplitNode(BaseTrialNodeMixin, DataClassNode):
    __tablename__ = 'trialsplit'
    id = Column(String, ForeignKey('dataclassnode.tree_id'), primary_key=True)
    split_number = Column(Integer, nullable=False, default=0)

    __mapper_args__ = {
        'polymorphic_identity': 'trialsplitdataclass',
    }

    def _update_attrs_from_dataclass(self, _dataclass: TrialSplitDataclass):
        BaseTrialNodeMixin._update_attrs_from_dataclass(self, _dataclass=_dataclass)
        self.split_number = _dataclass.split_number

    def to_empty_dataclass(self) -> TrialSplitDataclass:
        return TrialSplitDataclass(
            split_number=self.split_number, **self._attrs_for_to_dataclass())


class MetricResultsNode(DataClassNode):
    __tablename__ = 'metricresults'
    id = Column(String, ForeignKey('dataclassnode.tree_id'), primary_key=True)
    metric_name = Column(String, nullable=False, default=DEFAULT_METRIC_NAME)

    valid_values = relationship("ValidMetricResultsValues", back_populates="metric_results_node")
    train_values = relationship("TrainMetricResultsValues", back_populates="metric_results_node")

    higher_score_is_better = Column(Boolean, nullable=False, default=True)

    __mapper_args__ = {
        'polymorphic_identity': 'metricresultsdataclass',
    }

    def _update_attrs_from_dataclass(self, _dataclass: MetricResultsDataclass):
        self.metric_name = _dataclass.metric_name

        # Extending the lists with new values from metric results value tables.
        start: int = len(self.valid_values)
        self.valid_values.extend([
            ValidMetricResultsValues(value=v)
            for v in _dataclass.validation_values[start:]
        ])

        start: int = len(self.train_values)
        self.train_values.extend([
            TrainMetricResultsValues(value=v)
            for v in _dataclass.train_values[start:]
        ])

        self.higher_score_is_better = _dataclass.higher_score_is_better

    def to_empty_dataclass(self) -> MetricResultsDataclass:
        return MetricResultsDataclass(
            metric_name=self.metric_name,
            validation_values=[v.value for v in self.valid_values],
            train_values=[v.value for v in self.train_values],
            higher_score_is_better=self.higher_score_is_better
        )


@declarative_mixin
class MetricResultsValuesMixin:
    id = Column(Integer, Sequence('id_seq', metadata=Base.metadata), primary_key=True, index=True)

    value = Column(Float, nullable=False)
    # TODO: add datetime to metric results dataclass?
    datetime = Column(DateTime, default=func.now())

    def __init__(self, value: float, _datetime: datetime = None):
        self.value = value
        if _datetime is None:
            _datetime = datetime.now()
        self.datetime = _datetime


class TrainMetricResultsValues(MetricResultsValuesMixin, Base):
    __tablename__ = 'trainmetricresultsvalues'
    metric_results_node_id = Column(String, ForeignKey('dataclassnode.tree_id'))
    metric_results_node = relationship("MetricResultsNode", back_populates="train_values", uselist=False)


class ValidMetricResultsValues(MetricResultsValuesMixin, Base):
    __tablename__ = 'validmetricresultsvalues'
    metric_results_node_id = Column(String, ForeignKey('dataclassnode.tree_id'))
    metric_results_node = relationship("MetricResultsNode", back_populates="valid_values", uselist=False)


dataclass_2_sqlalchemy_node: typing.OrderedDict[Type[BaseDataclass], Type[DataClassNode]] = OrderedDict([
    (RootDataclass, DataClassNode),
    (ClientDataclass, ClientNode),
    (ProjectDataclass, ProjectNode),
    (RoundDataclass, RoundNode),
    (TrialDataclass, TrialNode),
    (TrialSplitDataclass, TrialSplitNode),
    (MetricResultsDataclass, MetricResultsNode),
])

sqlalchemy_node_2_dataclass: typing.OrderedDict[Type[BaseDataclass], Type[DataClassNode]] = OrderedDict([
    (v, k) for k, v in dataclass_2_sqlalchemy_node.items()
])


class AutoMLLog(Base):
    __tablename__ = 'automllog'
    id = Column(Integer, Sequence('id_seq', metadata=Base.metadata), primary_key=True, index=True)
    dataclass_node_id = Column(String, ForeignKey('dataclassnode.tree_id'))
    dataclass_node = relationship("DataClassNode", back_populates="logs", uselist=False)
    datetime = Column(DateTime, default=func.now())

    log_text = Column(TEXT)


class _DatabaseLoggerHandlerMixin:
    """
    Mixin to add a in-memory logging handler to a repository.
    """

    def add_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> 'HyperparamsRepository':
        # TODO: use AutoMLLog
        return self

    def get_log_from_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> str:
        # TODO: use AutoMLLog
        return logger.get_scoped_string_history()


class DatabaseHyperparamRepository(_DatabaseLoggerHandlerMixin, HyperparamsRepository):
    def __init__(self, engine, session):
        HyperparamsRepository.__init__(self)
        _DatabaseLoggerHandlerMixin.__init__(self)
        self.engine = engine
        self.session = session

    def create_db(self) -> 'DatabaseHyperparamRepository':
        Base.metadata.create_all(self.engine)
        _root_dc = RootDataclass()
        root_dcn = ScopedLocationTreeNode(DataClassNode(_root_dc), parent=None)
        self.session.add(root_dcn)
        self.session.commit()
        self.save(_root_dc, ScopedLocation(), deep=True)
        return self

    def load(self, scope: ScopedLocation, deep=False) -> SubDataclassT:

        # Get DataClassNode by parsing the tree with scopedlocation attributes:
        node: ScopedLocationTreeNode = ScopedLocationTreeNode.query(self.session, scope, deep=deep)

        if node is not None:
            return node.to_dataclass(deep=deep)
        else:
            return scope.new_dataclass_from_id()

    def save(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False) -> 'HyperparamsRepository':
        try:

            DataClassTreeNodeUpdater(self.session, scope, deep).add_or_update_dataclass(_dataclass, parent=None)

            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise e


class SQLLiteHyperparamsRepository(DatabaseHyperparamRepository):
    def __init__(self, sqllite_db_path, echo=False):

        sqlite_filepath = os.path.join(sqllite_db_path, "sqlite.db")
        os.makedirs(sqllite_db_path, exist_ok=True)
        engine = create_engine(f"sqlite:///{sqlite_filepath}", echo=echo, future=True)

        _Session = sessionmaker()
        _Session.configure(bind=engine)
        session = _Session()

        DatabaseHyperparamRepository.__init__(self, engine, session)
        self.create_db()


class PostGreSQLHyperparamsRepository(DatabaseHyperparamRepository):

    def __init__(self, postgresql_db_path, echo=False):
        raise NotImplementedError("TODO: implement this.")

    def get_database_path(self, user: str, password: str, host: str, dialect: str, driver: str = ''):
        """
        :param user:
        :param password:
        :param dialect: e.g. mysql
        :param driver: (Optional) e.g. "pymysql"
        :return:
        """
        # TODO: implement this.
        return "dialect[+driver]://user:password@host/dbname"



"""
Neuraxle's Flow Steps
====================================
Wrapper steps that apply some effect to the flow of data in the pipeline.

For instance, the following steps are wrappers of other steps and apply some
control logic to the flow of data into the steps:

- While,
- ForEach,
- BreakIf,
- ContinueIf,
- Try,
- Catch (Except),
- RaiseIf (Throw),
- If & ElseIf & Else,
- Switch

..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
from operator import attrgetter
from typing import Callable, Dict, Optional, Tuple, Union

from neuraxle.base import BaseStep, BaseTransformer
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import (ExecutionPhase, ForceHandleOnlyMixin, HandleOnlyMixin, MetaStep, NonFittableMixin,
                           TransformHandlerOnlyMixin, TruncableSteps)
from neuraxle.data_container import DataContainer as DACT
from neuraxle.data_container import ExpandedDataContainer
from neuraxle.hyperparams.distributions import Boolean, Choice
from neuraxle.hyperparams.space import HyperparameterSamples, HyperparameterSpace
from neuraxle.steps.numpy import NumpyConcatenateOnAxisIfNotEmpty
from neuraxle.union import FeatureUnion

OPTIONAL_ENABLED_HYPERPARAM = 'enabled'


class TrainOrTestOnlyWrapper(ForceHandleOnlyMixin, MetaStep):
    """
    A wrapper to run wrapped step only in test mode, or only in train mode.

    Execute only in test mode:

    .. code-block:: python

        p = TrainOrTestOnlyWrapper(Identity(), is_train_only=True)

    Execute only in train mode:

    .. code-block:: python

        p = TrainOnlyWrapper(Identity(), test_only=False)

    .. seealso::
        :class:`~neuraxle.steps.flow.TrainOnlyWrapper`,
        :class:`~neuraxle.steps.flow.TestOnlyWrapper`,
        :class:`~neuraxle.base.ForceHandleMixin`,
        :class:`~neuraxle.base.MetaStepMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, wrapped: BaseStep, is_train_only=True, cache_folder_when_no_handle=None):
        MetaStep.__init__(self, wrapped=wrapped)
        ForceHandleOnlyMixin.__init__(self, cache_folder=cache_folder_when_no_handle)

        self.is_train_only = is_train_only

    def _fit_data_container(self, data_container: DACT, context: CX) -> BaseStep:
        """
        :param data_container: data container
        :param context: execution context
        :return: step, data_container
        """
        if self._should_execute_wrapped_step():
            self.wrapped = self.wrapped.handle_fit(data_container, context)
            return self
        return self

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[BaseStep, DACT]:
        """
        :param data_container: data container
        :param context: execution context
        :return: step, data_container
        """
        if self._should_execute_wrapped_step():
            self.wrapped, data_container = self.wrapped.handle_fit_transform(data_container, context)
            return self, data_container
        return self, data_container

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        :param data_container: data container
        :param context: execution context
        :return: step, data_container
        """
        if self._should_execute_wrapped_step():
            return self.wrapped.handle_transform(data_container, context)
        return data_container

    def _should_execute_wrapped_step(self):
        return (self.wrapped.is_train and self.is_train_only) or (not self.wrapped.is_train and not self.is_train_only)


class TrainOnlyWrapper(TrainOrTestOnlyWrapper):
    """
    A wrapper to run wrapped step only in train mode

    Execute only in train mode:

    .. code-block:: python

        p = Pipeline([
            TrainOnlyWrapper(Identity())
        ])

    .. seealso::
        :class:`TrainOrTestOnlyWrapper`,
        :class:`TestOnlyWrapper`
    """

    def __init__(self, wrapped: BaseStep):
        TrainOrTestOnlyWrapper.__init__(self, wrapped=wrapped, is_train_only=True)


class TestOnlyWrapper(TrainOrTestOnlyWrapper):
    """
    A wrapper to run wrapped step only in test mode

    Execute only in train mode:

    .. code-block:: python

        p = Pipeline([
            TestOnlyWrapper(Identity())
        ])

    .. seealso::
        :class:`TrainOrTestOnlyWrapper`,
        :class:`TrainOnlyWrapper`
    """
    __test__ = False  # to avoid pytest to run this class as a test class.

    def __init__(self, wrapped: BaseStep):
        TrainOrTestOnlyWrapper.__init__(self, wrapped=wrapped, is_train_only=False)


class ExecuteIf(HandleOnlyMixin, MetaStep):
    def __init__(self, condition_function: Callable, wrapped: BaseStep):
        MetaStep.__init__(self, wrapped)
        HandleOnlyMixin.__init__(self)
        self.condition_function: Callable = condition_function

    def _fit_data_container(self, data_container: DACT, context: CX):
        if self.condition_function(self, data_container, context):
            return MetaStep._fit_data_container(self, data_container, context)
        return self

    def _fit_transform_data_container(self, data_container: DACT, context: CX):
        if self.condition_function(self, data_container, context):
            return MetaStep._fit_transform_data_container(self, data_container, context)
        return self, data_container

    def _transform_data_container(self, data_container: DACT, context: CX):
        if self.condition_function(self, data_container, context):
            return MetaStep._transform_data_container(self, data_container, context)
        return data_container


class IfExecutionPhaseIsThen(ExecuteIf):
    """
    If, at runtime, the execution phase is the same as the one given to the constructor, then execute wrapped step.

    By default, will raise an error if the execution phase is not specified in the context.
    Steps which implement ForceHandleMixin create context with unspecified phase on fit, fit_transform and transform call.
    """

    def __init__(self, phase: ExecutionPhase, wrapped: BaseTransformer, raise_if_phase_unspecified: bool = True):
        ExecuteIf.__init__(self, self.check_context, wrapped)
        self.phase = phase
        self.raise_if_phase_unspecified = raise_if_phase_unspecified

    def check_context(self, step, data_container, context: CX):
        if context.execution_phase == self.phase:
            return True
        elif self.raise_if_phase_unspecified and context.execution_phase == ExecutionPhase.UNSPECIFIED:
            raise ValueError("Execution phase is unspecified while a step requires it to be specified.")
        return False


class ExecutionPhaseSwitch(HandleOnlyMixin, TruncableSteps):
    def __init__(self, phase_to_callable: Dict[ExecutionPhase, BaseTransformer],
                 default: Optional[BaseTransformer] = None):
        phase, steps = zip(*phase_to_callable.items())
        if default:
            steps.append(default)
        TruncableSteps.__init__(self, steps_as_tuple=steps)
        self.phase_to_step_index = {p: i for i, p in enumerate(phase)}
        self.default = default

    def _get_step(self, context):
        if context.execution_phase not in self.phase_to_step_index.keys():
            if self.default is None:
                raise KeyError(f"No behaviour defined for {context.execution_phase}.")
            ind = -1
        else:
            ind = self.phase_to_step_index[context.execution_phase]
        return self.steps_as_tuple[ind][1]

    def _set_step(self, context, step):
        if context.execution_phase not in self.phase_to_step_index.keys():
            if self.default is None:
                raise KeyError(f"No behaviour defined for {context.execution_phase}.")
            ind = -1
        else:
            ind = self.phase_to_step_index[context.execution_phase]
        self.steps_as_tuple[ind] = (self.steps_as_tuple[ind][0], step)
        return self

    def _fit_data_container(self, data_container: DACT, context: CX) -> 'BaseStep':
        step = self._get_step(context).handle_fit(data_container, context)
        return self._set_step(context, step)

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseTransformer', DACT]:
        step, data_container = self._get_step(context).handle_fit_transform(data_container, context)
        return self._set_step(context, step), data_container

    def _transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseTransformer', DACT]:
        return self._get_step(context).handle_transform(data_container, context)


class OptionalStep(ForceHandleOnlyMixin, MetaStep):
    """
    A wrapper to nullify a step: nullify its hyperparams, and also nullify all of his behavior.
    Example usage:

    .. code-block:: python

        p = Pipeline([
            OptionalStep(Identity(), enabled=True)
        ])

    .. seealso::
        :class:`~neuraxle.base.MetaStep`,
        :class:`~neuraxle.metaopt.AutoML`,
    """

    def __init__(self, wrapped: BaseTransformer, enabled: bool = True, nullified_return_value=None,
                 cache_folder_when_no_handle=None, use_hyperparameter_space=True, nullify_hyperparams=True):
        hyperparameter_space = HyperparameterSpace({
            OPTIONAL_ENABLED_HYPERPARAM: Boolean()
        }) if use_hyperparameter_space else {}

        MetaStep.__init__(
            self,
            hyperparams=HyperparameterSamples({
                OPTIONAL_ENABLED_HYPERPARAM: enabled
            }),
            hyperparams_space=hyperparameter_space,
            wrapped=wrapped
        )
        ForceHandleOnlyMixin.__init__(self, cache_folder_when_no_handle)

        if nullified_return_value is None:
            nullified_return_value = []
        self.nullified_return_value = nullified_return_value
        self.nullify_hyperparams = nullify_hyperparams

    def _fit_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseTransformer', DACT]:
        """
        Nullify wrapped step hyperparams, and don't fit the wrapped step.

        :param data_container: data container
        :param context: execution context
        :return: step, data_container
        """
        if self.hyperparams[OPTIONAL_ENABLED_HYPERPARAM]:
            self.wrapped = self.wrapped.handle_fit(data_container, context)
            return self
        else:
            self._nullify_hyperparams()
            return self

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseTransformer', DACT]:
        """
        Nullify wrapped step hyperparams, and don't fit_transform the wrapped step.

        :param data_container: data container
        :param context: execution context
        :return: step, data_container
        """
        if self.hyperparams[OPTIONAL_ENABLED_HYPERPARAM]:
            self.wrapped, data_container = self.wrapped.handle_fit_transform(data_container, context)
            return self, data_container
        else:
            self._nullify_hyperparams()
            return self, self._passtrough_dact(data_container)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Nullify wrapped step hyperparams, and don't transform the wrapped step.

        :param data_container: data container
        :param context: execution context
        :return: step, data_container
        """
        if self.hyperparams[OPTIONAL_ENABLED_HYPERPARAM]:
            return self.wrapped.handle_transform(data_container, context)
        else:
            self._nullify_hyperparams()
            return self._passtrough_dact(data_container)

    def _nullify_hyperparams(self):
        """
        Nullify wrapped step hyperparams using hyperparams_space.nullify().
        """
        if not self.nullify_hyperparams:
            return
        hyperparams_space = self.wrapped.get_hyperparams_space()
        self.wrapped.set_hyperparams(hyperparams_space.nullify())

    def _passtrough_dact(self, data_container):
        return DACT(
            data_inputs=self.nullified_return_value,
            ids=data_container.ids,
            expected_outputs=self.nullified_return_value
        )


class ChooseStepElseIdentity(OptionalStep):
    def __init__(self, wrapped: BaseTransformer, enabled: bool = True, nullify_hyperparams=True):
        OptionalStep.__init__(self, wrapped, enabled, None, None, True, nullify_hyperparams)

    def _passtrough_dact(self, data_container):
        return data_container


class ChooseOneStepOf(FeatureUnion):
    CHOICE_HYPERPARAM = 'choice'
    """
    A pipeline to allow choosing one step using an hyperparameter.

    Example usage :

    .. code-block:: python

        p = Pipeline([
            ChooseOneStepOf([
                ('a', Identity()),
                ('b', Identity())
            ])
        ])
        p.set_hyperparams({
            'ChooseOneOrManyStepsOf__choice': 'a',
        })
        # or
        p.set_hyperparams({
            'ChooseOneStepOf': {
                'a': { 'enabled': True },
                'b': { 'enabled': False }
            }
        })

    .. seealso::
        :class:`~neuraxle.pipeline.Pipeline`
        :class:`Optional`
    """

    def __init__(self, steps, default_choice=None):
        FeatureUnion.__init__(self, steps, joiner=SelectNonEmptyDataContainer())

        self._make_all_steps_optional()

        choices = list(self.keys())[:-1]

        if default_choice is None:
            self.update_hyperparams({
                ChooseOneStepOf.CHOICE_HYPERPARAM: choices[0]
            })
        else:
            self.update_hyperparams({
                ChooseOneStepOf.CHOICE_HYPERPARAM: default_choice
            })
        self.update_hyperparams_space({
            ChooseOneStepOf.CHOICE_HYPERPARAM: Choice(choices)
        })

    def set_hyperparams(self, hyperparams: Union[HyperparameterSamples, dict]):
        super().set_hyperparams(hyperparams)
        self._update_optional_hyperparams()
        return self

    def _set_hyperparams(self, hyperparams: HyperparameterSamples) -> HyperparameterSamples:
        ret = super()._set_hyperparams(hyperparams)
        self._update_optional_hyperparams()
        return ret

    def update_hyperparams(self, hyperparams: Union[Dict, HyperparameterSamples]) -> 'BaseTransformer':
        super().update_hyperparams(hyperparams)
        self._update_optional_hyperparams()
        return self

    def _update_hyperparams(self, hyperparams: Union[HyperparameterSamples, dict]):
        """
        Set chosen step hyperparams.

        :param hyperparams: hyperparams
        :type hyperparams: HyperparameterSamples
        :return:
        """
        ret = super()._update_hyperparams(hyperparams)
        self._update_optional_hyperparams()
        return ret

    def _update_optional_hyperparams(self):
        step_names = list(self.keys())
        chosen_step_name = self.hyperparams[self.CHOICE_HYPERPARAM] if self.CHOICE_HYPERPARAM in self.hyperparams \
            else step_names[0]

        if chosen_step_name not in step_names:
            raise ValueError('Invalid Chosen Step in {0}'.format(self.name))
        for step_name in step_names[:-1]:
            if step_name == chosen_step_name:
                self[chosen_step_name].set_hyperparams({
                    OPTIONAL_ENABLED_HYPERPARAM: True
                })
            else:
                self[step_name].set_hyperparams({
                    OPTIONAL_ENABLED_HYPERPARAM: False
                })

    def _make_all_steps_optional(self):
        """
        Wrap all steps with :class:`Optional` wrapper.
        """
        step_names = list(self.keys())
        for step_name in step_names[:-1]:
            self[step_name] = OptionalStep(
                self[step_name].set_name('OptionalStep({})'.format(step_name)),
                use_hyperparameter_space=False,
                nullify_hyperparams=False
            )

        self._refresh_steps()


class ChooseOneOrManyStepsOf(FeatureUnion):
    """
    A pipeline to allow choosing many steps using an hyperparameter.

    Example usage :

    .. code-block:: python

        p = Pipeline([
            ChooseOneOrManyStepsOf([
                ('a', Identity()),
                ('b', Identity())
            ])
        ])
        p.set_hyperparams({
            'ChooseOneOrManyStepsOf__a__enabled': True,
            'ChooseOneOrManyStepsOf__b__enabled': False
        })
        # or
        p.set_hyperparams({
            'ChooseOneOrManyStepsOf': {
                'a': { 'enabled': True },
                'b': { 'enabled': False }
            }
        })

    .. seealso::
        :class:`~neuraxle.pipeline.Pipeline`
        :class:`Optional`
    """

    def __init__(self, steps, joiner: NonFittableMixin = None):
        if joiner is None:
            joiner = NumpyConcatenateOnAxisIfNotEmpty(axis=-1)
        FeatureUnion.__init__(self, steps_as_tuple=steps, joiner=joiner)
        self.set_hyperparams(HyperparameterSamples({}))
        self._make_all_steps_optional()

    def _make_all_steps_optional(self):
        """
        Wrap all steps with :class:`Optional` wrapper.
        """
        step_names = list(self.keys())
        for step_name in step_names[:-1]:
            self[step_name] = OptionalStep(self[step_name])
        self._refresh_steps()


class SelectNonEmptyDataInputs(TransformHandlerOnlyMixin, BaseTransformer):
    """
    A step that selects non empty data inputs.

    .. seealso::
        :class:`~neuraxle.base.TransformHandlerOnlyMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self):
        BaseTransformer.__init__(self)
        TransformHandlerOnlyMixin.__init__(self)

    def _transform_data_container(self, data_container: DACT, context: CX):
        """
        Handle transform.

        :param data_container: the data container to join
        :param context: execution context
        :return: transformed data container
        """
        data_inputs = [dc.data_inputs for dc in data_container.data_inputs if len(dc.data_inputs) > 0]
        if len(data_inputs) == 1:
            data_inputs = data_inputs[0]

        data_container = DACT(data_inputs=data_inputs, ids=data_container.ids,
                              expected_outputs=data_container.expected_outputs)

        return data_container


class SelectNonEmptyDataContainer(TransformHandlerOnlyMixin, BaseTransformer):
    """
    A step that selects non empty data containers.
    Assumes that the given DataContainer contains a list of DataContainer as data_inputs.

    .. seealso::
        :class:`~neuraxle.base.TransformHandlerOnlyMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self):
        BaseTransformer.__init__(self)
        TransformHandlerOnlyMixin.__init__(self)

    def _transform_data_container(self, data_container: DACT, context: CX):

        filtered_data_containers = list(filter(
            lambda dc: (len(dc.data_inputs) > 0 or len(dc.expected_outputs) > 0),
            data_container.data_inputs
        ))
        if len(filtered_data_containers) == 1:
            return filtered_data_containers[0]
        else:
            return DACT(
                ids=data_container._ids,
                di=list(map(attrgetter("data_inputs"))),
                eo=list(map(attrgetter("expected_outputs"))),
            )


class ExpandDim(MetaStep):
    """
    Similar to numpys expand_dim function, ExpandDim step expands the dimension of all the data inside the data container.
    ExpandDim sends the expanded data container to the wrapped step.

    This is akin from passing the dact data from `shape` to `[1, *shape]` within the wrapped step,
    to then by default go back to the original shape (optional). The ids will now contain a summary id temporarily.

    ExpandDim returns the transformed expanded dim reduced to its original shape (see :func:`~neuraxle.steps.loop.ExpandedDataContainer.reduce_dim`).

    The wrapped step will receive a single current_id, data_input, and expected output:
        - The ids is a list of one element that contains a single summary id created from all of the current ids.
        - The data_inputs is a list of one element that contains the original expected outputs list.
        - The expected_outputs is a list of one element that contains the original expected outputs list.

    .. seealso::
        :class:`~neuraxle.data_container.ExpandedDataContainer`
    """

    def __init__(self, wrapped: BaseTransformer, then_unexpand: bool = True):
        MetaStep.__init__(self, wrapped)

        self.then_unexpand: bool = then_unexpand

    def _will_process(self, data_container: DACT, context: CX) -> Tuple[ExpandedDataContainer, CX]:
        data_container, context = BaseStep._will_process(self, data_container, context)
        return ExpandedDataContainer.create_from(data_container), context

    def _did_process(self, data_container: ExpandedDataContainer, context: CX) -> DACT:
        data_container: ExpandedDataContainer = super()._did_process(data_container, context)
        if self.then_unexpand:
            data_container = data_container.reduce_dim()
        return data_container


class ReversiblePreprocessingWrapper(HandleOnlyMixin, TruncableSteps):
    """
    TruncableSteps with a preprocessing step(1), and a postprocessing step(2)
    that inverse transforms with the preprocessing step at the end (1, 2, reversed(1)).

    Example usage :

    .. code-block:: python

        step = ReversiblePreprocessingWrapper(
            preprocessing_step=MultiplyBy2(),
            postprocessing_step=Add10()
        )

        outputs = step.transform(np.array(range(5)))

        assert np.array_equal(outputs, np.array([5, 6, 7, 8, 9]))

    """

    def __init__(self, preprocessing_step, postprocessing_step):
        TruncableSteps.__init__(self, [
            ("preprocessing_step", preprocessing_step),
            ("postprocessing_step", postprocessing_step)
        ])
        HandleOnlyMixin.__init__(self)

    def _fit_data_container(self, data_container: DACT,
                            context: CX) -> 'ReversiblePreprocessingWrapper':
        """
        Handle fit by fitting preprocessing step, and postprocessing step.

        :param data_container: data container to fit on
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: self, data_container
        :rtype: (ReversiblePreprocessingWrapper, DataContainer)
        """
        self["preprocessing_step"], data_container = self["preprocessing_step"].handle_fit_transform(
            data_container, context)
        self["postprocessing_step"] = self["postprocessing_step"].handle_fit(
            data_container, context)

        return self

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        According to the idiom of `(1, 2, reversed(1))`, we do this, in order:

            - `1`. Transform preprocessing step
            - `2`. Transform postprocessing step
            - `reversed(1)`. Inverse transform preprocessing step

        :param data_container: data container to transform
        :param context: execution context
        :return: data_container
        """
        data_container = self["preprocessing_step"].handle_transform(
            data_container, context.push(self["preprocessing_step"]))
        data_container = self["postprocessing_step"].handle_transform(
            data_container, context.push(self["postprocessing_step"]))
        data_container = self["preprocessing_step"].handle_inverse_transform(
            data_container, context.push(self["preprocessing_step"]))

        return data_container

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[BaseStep, DACT]:
        """
        According to the idiom of `(1, 2, reversed(1))`, we do this, in order:

            - `1`. Fit Transform preprocessing step
            - `2`. Fit Transform postprocessing step
            - `reversed(1)`. Inverse transform preprocessing step

        :param data_container: data container to transform
        :param context: execution context
        :return: (self, data_container)
        """
        self["preprocessing_step"], data_container = self["preprocessing_step"].handle_fit_transform(
            data_container, context.push(self["preprocessing_step"]))
        self["postprocessing_step"], data_container = self["postprocessing_step"].handle_fit_transform(
            data_container, context.push(self["postprocessing_step"]))
        data_container = self["preprocessing_step"].handle_inverse_transform(
            data_container, context.push(self["preprocessing_step"]))

        return self, data_container



"""
Pipeline Steps Based on Scikit-Learn
=====================================
Those steps works with scikit-learn (sklearn) transformers and estimators.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import functools
import inspect
from typing import Any, Tuple

from neuraxle.base import BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.hyperparams.distributions import Boolean, LogUniform
from neuraxle.hyperparams.space import (HyperparameterSamples,
                                        HyperparameterSpace, RecursiveDict)
from neuraxle.steps.numpy import NumpyTranspose
from neuraxle.union import ModelStacking

from sklearn.base import BaseEstimator
from sklearn.ensemble import BaseEnsemble
from sklearn.linear_model import Ridge


class SKLearnWrapper(BaseStep):
    def __init__(
            self,
            wrapped_sklearn_predictor,
            hyperparams_space: HyperparameterSpace = None,
            return_all_sklearn_default_params_on_get: bool = False,
            use_partial_fit: bool = False,
            use_predict_proba: bool = False,
            partial_fit_kwargs: dict = None
    ):
        if not isinstance(wrapped_sklearn_predictor, BaseEstimator):
            raise ValueError("The wrapped_sklearn_predictor must be an instance of scikit-learn's BaseEstimator.")
        self.wrapped_sklearn_predictor = wrapped_sklearn_predictor
        self.is_ensemble = isinstance(wrapped_sklearn_predictor, BaseEnsemble)
        params: dict = wrapped_sklearn_predictor.get_params()
        self._delete_base_estimator_from_dict(params)
        BaseStep.__init__(self, hyperparams=params, hyperparams_space=hyperparams_space)
        self.return_all_sklearn_default_params_on_get = return_all_sklearn_default_params_on_get
        self.name += "_" + wrapped_sklearn_predictor.__class__.__name__
        self.use_partial_fit: bool = use_partial_fit
        if self.use_partial_fit:
            if partial_fit_kwargs is None:
                partial_fit_kwargs = {}
            self.partial_fit_kwargs = partial_fit_kwargs
        self.use_predict_proba: bool = use_predict_proba

    def _setup(self, context: CX = None) -> 'SKLearnWrapper':
        BaseStep._setup(self, context)
        if self.use_partial_fit:
            self.wrapped_sklearn_predictor.fit = functools.partial(self.wrapped_sklearn_predictor.partial_fit,
                                                                   **self.partial_fit_kwargs)
        return self

    def _delete_base_estimator_from_dict(self, params):
        """
        Sklearn BaseEnsemble models contain other models as parameter; those can't be json encoded. We retrieve the parameters of theses sub-models on a .get_params(deep=True) call, we simply need to delete them from the parameter dictionary to avoid errors when saving/loading hyperparameters.
        """
        for name in list(params.keys()):
            if isinstance(params[name], BaseEstimator):
                del params[name]

    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Any]:
        if hasattr(self.wrapped_sklearn_predictor, 'fit_transform'):
            if expected_outputs is None or len(inspect.getfullargspec(self.wrapped_sklearn_predictor.fit).args) < 3:
                out = self._sklearn_fit_transform_without_expected_outputs(data_inputs)
            else:
                out = self._sklearn_fit_transform_with_expected_outputs(data_inputs, expected_outputs)
            return self, out

        self.fit(data_inputs, expected_outputs)

        return self, self.transform(data_inputs)

    def _sklearn_fit_transform_with_expected_outputs(self, data_inputs, expected_outputs):
        self.wrapped_sklearn_predictor = self.wrapped_sklearn_predictor.fit(data_inputs, expected_outputs)
        return self.transform(data_inputs)

    def _sklearn_fit_transform_without_expected_outputs(self, data_inputs):
        if self.use_partial_fit:
            self.wrapped_sklearn_predictor = self.wrapped_sklearn_predictor.fit(data_inputs)
            out = self.transform(data_inputs)
        else:
            out = self.wrapped_sklearn_predictor.fit_transform(data_inputs)
        return out

    def fit(self, data_inputs, expected_outputs=None) -> 'SKLearnWrapper':
        if expected_outputs is None or len(inspect.getfullargspec(self.wrapped_sklearn_predictor.fit).args) < 3:
            self._sklearn_fit_without_expected_outputs(data_inputs)
        else:
            self._sklearn_fit_with_expected_outputs(data_inputs, expected_outputs)
        return self

    def _sklearn_fit_with_expected_outputs(self, data_inputs, expected_outputs):
        self.wrapped_sklearn_predictor = self.wrapped_sklearn_predictor.fit(data_inputs, expected_outputs)

    def _sklearn_fit_without_expected_outputs(self, data_inputs):
        self.wrapped_sklearn_predictor = self.wrapped_sklearn_predictor.fit(data_inputs)

    def transform(self, data_inputs):
        if self.use_predict_proba and hasattr(self.wrapped_sklearn_predictor, 'predict_proba'):
            return self.wrapped_sklearn_predictor.predict_proba(data_inputs)
        elif hasattr(self.wrapped_sklearn_predictor, 'predict'):
            return self.wrapped_sklearn_predictor.predict(data_inputs)
        return self.wrapped_sklearn_predictor.transform(data_inputs)

    def _set_hyperparams(self, hyperparams: HyperparameterSamples) -> BaseStep:
        """
        Set hyperparams for base step, and the wrapped sklearn_predictor.

        :param hyperparams:
        :return: self
        """
        # set the step hyperparams, and set the wrapped sklearn predictor params
        BaseStep._set_hyperparams(self, hyperparams)
        self.wrapped_sklearn_predictor.set_params(
            **hyperparams.with_separator(RecursiveDict.DEFAULT_SEPARATOR).to_flat_dict()
        )

        return self.hyperparams

    def _update_hyperparams(self, hyperparams: HyperparameterSamples) -> BaseStep:
        """
        Update hyperparams for base step, and the wrapped sklearn_predictor.

        :param hyperparams:
        :return: self
        """
        # update the step hyperparams, and update the wrapped sklearn predictor params
        BaseStep._update_hyperparams(self, hyperparams)
        self.wrapped_sklearn_predictor.set_params(
            **self.hyperparams.with_separator(RecursiveDict.DEFAULT_SEPARATOR).to_flat_dict()
        )

        return self.hyperparams

    def _get_hyperparams(self):
        if self.return_all_sklearn_default_params_on_get:
            hp = self.wrapped_sklearn_predictor.get_params()
            self._delete_base_estimator_from_dict(hp)
            return HyperparameterSamples(hp)
        else:
            return BaseStep._get_hyperparams(self)

    def get_wrapped_sklearn_predictor(self):
        return self.wrapped_sklearn_predictor

    def _repr(self, level=0, verbose=False) -> str:
        type_ = self.__class__
        module = type_.__module__
        qualname = type_.__qualname__
        wrappedname = str(self.wrapped_sklearn_predictor.__class__.__name__)
        return "<{}.{}({}(...)) object {}>".format(module, qualname, wrappedname, hex(id(self)))


class RidgeModelStacking(ModelStacking):
    def __init__(self, brothers):
        ModelStacking.__init__(
            self,
            brothers,
            SKLearnWrapper(
                Ridge(),
                HyperparameterSpace({"alpha": LogUniform(0.1, 10.0), "fit_intercept": Boolean()})
            ),
            joiner=NumpyTranspose()
        )



"""
Neuraxle's Hyperparameter Optimizer Base Classes
====================================================

Not all hyperparameter optimizers are there, but the base can be found here.


.. seealso::
    :class:`~neuraxle.metaopt.hyperopt.tpe.TreeParzenEstimator`,


..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""

import copy
import itertools
import math
import operator
import random
import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict
from functools import reduce
from typing import Any, List, Set, Tuple

from neuraxle.hyperparams.distributions import (
    ContinuousHyperparameterDistribution, DiscreteHyperparameterDistribution)
from neuraxle.hyperparams.space import (FlatDict, HyperparameterSamples,
                                        HyperparameterSpace)
from neuraxle.metaopt.data.reporting import RoundReport


class BaseHyperparameterOptimizer(ABC):

    @abstractmethod
    def find_next_best_hyperparams(self, _round: RoundReport, hp_space: HyperparameterSpace) -> HyperparameterSamples:
        """
        Find the next best hyperparams using previous trials, that is the
        whole :class:`neuraxle.metaopt.data.aggregate.Round`.

        :param round: a :class:`neuraxle.metaopt.data.aggregate.Round`
        :return: next hyperparameter samples to train on
        """
        raise NotImplementedError()


class HyperparameterSamplerStub(BaseHyperparameterOptimizer):

    def __init__(self, preconfigured_hp_samples: HyperparameterSamples):
        self.preconfigured_hp_samples = preconfigured_hp_samples

    def find_next_best_hyperparams(self, _round: RoundReport, hp_space: HyperparameterSpace) -> HyperparameterSamples:
        return self.preconfigured_hp_samples


class RandomSearchSampler(BaseHyperparameterOptimizer):
    """
    AutoML Hyperparameter Optimizer that randomly samples the space of random variables.
    Please refer to :class:`AutoML` for a usage example.

    .. seealso::
        :class:`Trainer`,
        :class:`HyperparamsRepository`,
    """

    def __init__(self):
        BaseHyperparameterOptimizer.__init__(self)

    def find_next_best_hyperparams(self, _round: RoundReport, hp_space: HyperparameterSpace) -> HyperparameterSamples:
        """
        Randomly sample the next hyperparams to try.

        :param round: round report
        :return: next random hyperparams
        """
        return hp_space.rvs()


class GridExplorationSampler(BaseHyperparameterOptimizer):
    """
    This hyperparameter space optimizer is similar to a grid search, however, it does
    try to greedily sample maximally different points in the space to explore it. This space
    optimizer has a fixed pseudorandom exploration method that makes the sampling reproductible.

    When over the expected_n_trials (if sampling too much), the sampler will turn
    to a non-seeded random search.

    It may be good for space exploration before a TPE or for unit tests.

    If the expected_n_trials is not set or set to 0, the sampler will guess its ideal
    sampling count and then switch to random search after that.
    """

    def __init__(self, expected_n_trials: int = 0, seed_i: int = 0):
        BaseHyperparameterOptimizer.__init__(self)
        self.expected_n_trials: int = expected_n_trials
        self._i: int = seed_i

    @staticmethod
    def estimate_ideal_n_trials(hp_space: HyperparameterSpace) -> int:
        _ges: GridExplorationSampler = GridExplorationSampler(expected_n_trials=1)
        _ges._reinitialize_grid(hp_space, [])

        _expected_n_trials = _ges.expected_n_trials
        # Then readjust the expected_n_trials to be a multiple of the number of hyperparameters:
        #     NOTE: TRIED A DOZEN OF POWS SQRT CUMSUMPROD AND SOPHISTICATED OTHER THINGS, AND FOUND THIS BEST ONE:
        _estimated_ideal_n_trials: int = math.ceil(0.29 * _expected_n_trials)
        # TODO: could add more samples to the grid lens to match the counts in _estimated_ideal_n_trials.

        return _estimated_ideal_n_trials

    def _reinitialize_grid(self, hp_space: HyperparameterSpace, previous_trials_hp: List[HyperparameterSamples]) -> HyperparameterSamples:
        """
        Update the grid exploration sampler.

        :param round_scope: round scope
        :return: next random hyperparams
        """
        self._n_sampled = 0
        self.flat_hp_grid_values: OrderedDict[str, List[Any]] = OrderedDict()
        self.flat_hp_grid_lens: List[int] = []
        # TODO: could make use of a ND array here to keep track of the grid exploration instead of using random too much. And a walk method picking the most L2-distant point, permutated over the last mod 3 samples to walk awkwardly like [mid, begin, fartest, side, other side] in the ND cube, also avoiding same-seen values.
        self._seen_hp_grid_values: Set[Tuple[int]] = set()

        self._generate_grid(hp_space)
        for flat_dict_sample in previous_trials_hp:
            self._reshuffle_grid()

            vals: Tuple[int] = tuple(flat_dict_sample.values())
            self._seen_hp_grid_values.add(vals)

    def find_next_best_hyperparams(self, _round: RoundReport, hp_space: HyperparameterSpace) -> HyperparameterSamples:
        """
        Sample the next hyperparams to try.

        :param round_scope: round scope
        :return: next hyperparams
        """
        self._reinitialize_grid(hp_space, _round.get_all_hyperparams())

        _space_max = reduce(operator.mul, self.flat_hp_grid_lens, 1)

        if self._n_sampled >= max(self.expected_n_trials, _space_max):
            return RandomSearchSampler().find_next_best_hyperparams(_round, hp_space)
        for _ in range(_space_max):
            i_grid_keys: Tuple[int] = tuple(self._gen_keys_for_grid())
            grid_values: OrderedDict[str, Any] = tuple(self[i_grid_keys].values())
            if grid_values in self._seen_hp_grid_values:
                self._reshuffle_grid()
                self._i += 1
            else:
                break

        # Second chance at picking unseen value yet.
        if grid_values in self._seen_hp_grid_values:
            # overwrite it:
            prod_values = list(itertools.product(*self.flat_hp_grid_values.values()))
            generator = random.Random(self._i)
            generator.shuffle(prod_values)
            for grid_values in prod_values:
                if grid_values not in self._seen_hp_grid_values:
                    break
            if grid_values in self._seen_hp_grid_values:
                grid_values = generator.choice(prod_values)

        # assert grid_values not in self._seen_hp_grid_values  # TODO: TMP.

        # value is finally chosen:
        self._seen_hp_grid_values.add(grid_values)
        _full_dict = OrderedDict([(a, b) for a, b in zip(
            self.flat_hp_grid_values.keys(),
            grid_values
        )])
        return HyperparameterSamples(_full_dict)

    def _generate_grid(self, hp_space: HyperparameterSpace):
        """
        Generate the grid of hyperparameters to pick from.

        :param hp_space: hyperparameter space
        """
        # Start with discrete params:
        for hp_name, hp_dist in hp_space.to_flat_dict().items():
            if isinstance(hp_dist, DiscreteHyperparameterDistribution):
                hp_samples: List[Any] = hp_dist.values()

                reordered_hp_samples = self._pseudo_shuffle_list(hp_samples)
                self.flat_hp_grid_values[hp_name] = reordered_hp_samples
                self.flat_hp_grid_lens.append(len(reordered_hp_samples))

        # Then fill the remaining continous params using the expected_n_trials:
        remainder: int = max(3, self.expected_n_trials)
        for hp_name, hp_dist in hp_space.to_flat_dict().items():
            if isinstance(hp_dist, ContinuousHyperparameterDistribution):
                hp_samples: List[Any] = [
                    hp_dist.mean(),
                    hp_dist.min() if hp_dist.is_discrete() else hp_dist._pseudo_min(),
                    hp_dist.max() if hp_dist.is_discrete() else hp_dist._pseudo_max(),
                    hp_dist.mean() + hp_dist.std(),
                    hp_dist.mean() - hp_dist.std(),
                    hp_dist.mean() + hp_dist.std() / 2,
                    hp_dist.mean() - hp_dist.std() / 2,
                    hp_dist.mean() + hp_dist.std() * 1.5,
                    hp_dist.mean() - hp_dist.std() * 1.5,
                    hp_dist.mean() + hp_dist.std() / 4,
                    hp_dist.mean() - hp_dist.std() / 4,
                    hp_dist.mean() + hp_dist.std() * 2.5,
                    hp_dist.mean() - hp_dist.std() * 2.5,
                ]

                def _ensure_unique(value: Any, _unique_set: Set[Any]) -> bool:
                    # remove duplicates such as when (or if) STD is of 0.
                    try:
                        ret = value not in _unique_set
                        _unique_set.add(value)
                        return ret
                    except BaseException:
                        return True
                _unique_set: Set[Any] = set()
                hp_samples: List[Any] = [
                    x for x in hp_samples[:remainder]
                    if x in hp_dist and not (math.isinf(x) or math.isnan(x)) and _ensure_unique(x, _unique_set)
                ]
                self.flat_hp_grid_values[hp_name] = hp_samples
                self.flat_hp_grid_lens.append(len(hp_samples))

        _estimated_ideal_n_trials: int = sum(self.flat_hp_grid_lens)
        _space_max = reduce(operator.mul, self.flat_hp_grid_lens, 1)

        if not (_estimated_ideal_n_trials <= self.expected_n_trials <= _space_max):
            _new_val = max(min(_space_max, self.expected_n_trials), _estimated_ideal_n_trials)
            warnings.warn(
                f"Warning: changed {self.__class__.__name__}.expected_n_trials from "
                f"{self.expected_n_trials} to {_new_val}. RandomSearch will be used "
                f"as a fallback past this point if needed.")
            self.expected_n_trials = _new_val

        self._i = 1

    @staticmethod
    def _pseudo_shuffle_list(x: List, seed: int = 0) -> list:
        """
        Shuffle a list to create a pseudo-random order that is interesting.
        """
        x = copy.copy(x)
        for i in reversed(range(len(x))):
            v = x[i]
            if (len(x) + i + seed) % 2 and (i + seed) != 0:
                del x[i]
                x.insert(1 - (seed % 2), v)
        return x

    def _gen_keys_for_grid(self) -> List[int]:
        """
        Generate the keys for the grid.

        :param i: index
        :return: keys
        """
        flat_idx: List[int] = []
        for _len in self.flat_hp_grid_lens:
            flat_idx.append(self._i % _len)
        return flat_idx

    def _reshuffle_grid(self, new_sample: FlatDict = None):
        """
        Reshuffling with pseudo-random seed the hyperparameters' values:
        """
        assert self._i != 0, "Cannot reshuffle the grid when _i is 0. Please generate grid first and increase _i."

        for seed, (k, v) in enumerate(self.flat_hp_grid_values.items()):
            if self._i % len(v) == 0:
                # reshuffling the grid for when it comes to the end of each of its sublist.
                # TODO: make a test for this to ensure that over an infinity of trials that the lists are shuffled evenly, or use a real and repeatable pseudorandom rng generator for it.
                new_v = self._pseudo_shuffle_list(v, seed % (self._i + 1) + self._i % (seed + 1))
                self.flat_hp_grid_values[k] = new_v

                if (seed + self._i / len(v)) % 3 == 0:
                    # reversing the grid at each 3 reshuffles + seed as it may sometimes skip the zero in the disorder.
                    self.flat_hp_grid_values[k] = list(reversed(self.flat_hp_grid_values[k]))
                    self._n_sampled += 1

        self._i += 1

    def __getitem__(self, i_grid_keys: List[int]) -> 'OrderedDict[str, Any]':
        """
        Access the keys for the grid.

        :param i_grid_keys: keys
        :return: hyperparams
        """
        flat_result: OrderedDict[str, Any] = OrderedDict()
        for j, hp_name in enumerate(self.flat_hp_grid_values.keys()):
            flat_result[hp_name] = self.flat_hp_grid_values[hp_name][i_grid_keys[j]]
        return flat_result



"""
Neuraxle's AutoML Metric Reporting classes.
=====================================================

Classes are splitted like this for the metric analysis:

- ProjectReport
- ClientReport
- RoundReport
- TrialReport
- TrialSplitReport

These AutoML reports are used to get information from the nested dataclasses, such as to create visuals.

..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
import json
import pprint
import typing
from collections import OrderedDict, defaultdict
from numbers import Number
from typing import (Any, Dict, Generic, Iterable, List, Optional, Tuple, Type,
                    TypeVar, Union)

import numpy as np
import pandas as pd
from neuraxle.base import TrialStatus
from neuraxle.hyperparams.space import (FlatDict, HyperparameterSamples,
                                        RecursiveDict)
from neuraxle.metaopt.data.vanilla import (BaseDataclass, ClientDataclass,
                                           MetricResultsDataclass,
                                           ProjectDataclass, RootDataclass,
                                           RoundDataclass, ScopedLocationAttr,
                                           ScopedLocationAttrInt,
                                           SubDataclassT, TrialDataclass,
                                           TrialSplitDataclass,
                                           dataclass_2_id_attr, from_json,
                                           to_json)

SubReportT = TypeVar('SubReportT', bound=Optional['BaseReport'])


def _filter_df_hps(wildcarded_hps: FlatDict, wildcards_to_keep: List[str] = None) -> FlatDict:
    """
    Filters the hyperparameters so as to keep only the indicated wildcards to keep.
    Also parses the hyperparameters to strings if not numeric.
    """
    def _to_str_if_not_number(value: Any) -> Union[str, Number]:
        if isinstance(value, Number) and not isinstance(value, bool):
            return value
        if isinstance(value, str):
            return repr(value)
        return str(value)
    return {
        k: _to_str_if_not_number(v) for k, v in wildcarded_hps.items()
        if wildcards_to_keep is None or k in wildcards_to_keep
    }


class BaseReport(Generic[SubReportT, SubDataclassT]):
    """
    A report contains a dataclass of the same subclass-level of itself, so as to be able to
    dig into the dataclass so as to observe it, such as to generate statistics and query its information.
    The dataclasses represent the results of an AutoML optimization round, even multiple rounds.
    These AutoML reports are used to get information from the nested dataclasses, such as to create visuals.

    .. seealso::
        :class:`neuraxle.metaopt.data.vanilla.BaseDataclass`
        :class:`neuraxle.metaopt.data.reporting.BaseAggregate`
    """
    def __init__(self, dc: SubDataclassT):
        self._dataclass: SubDataclassT = dc

    @staticmethod
    def from_dc(dc: SubDataclassT) -> SubReportT:
        return dataclass_2_report[dc.__class__](dc)

    @staticmethod
    def from_json(json_dc_data: str) -> SubReportT:
        return BaseReport.from_dc(from_json(json_dc_data))

    def get_id(self) -> ScopedLocationAttr:
        return self._dataclass.get_id()

    def info_df(self) -> pd.DataFrame:
        _json: str = to_json(self._dataclass.empty())

        info: dict = json.loads(_json)
        del info['__type__']
        del info[self._dataclass._sublocation_attr_name]
        del info[self._dataclass._id_attr_name]
        info = {k: pprint.pformat(v) if not isinstance(v, dict) else pprint.pformat(
            dict(HyperparameterSamples(v).to_flat_dict(use_wildcards=True))) for k, v in info.items()}

        column_names = ['Attribute', 'Value']
        if len(info) == 0:
            df = pd.DataFrame(columns=column_names)
        else:
            df = pd.DataFrame.from_records([info], columns=list(info.keys()))
            df = df.transpose()
            df = df.reset_index(level=0)
            df.columns = column_names
        return df

    def __len__(self) -> int:
        return len(self._dataclass.get_sublocation())

    def __iter__(self) -> Iterable[SubReportT]:
        for subdataclass in self._dataclass.get_sublocation_values():
            if subdataclass is not None:
                yield self.subreport(subdataclass)

    def __getitem__(self, item: ScopedLocationAttr) -> SubReportT:
        """
        Get trial at the given index.

        :param item: trial index
        :return:
        """
        subdataclass = self._dataclass.get_sublocation()[item]
        return self.subreport(subdataclass)

    def subreport(self, _dataclass: SubDataclassT) -> SubReportT:
        return report_2_subreport[self.__class__](_dataclass)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        subobjs = str([str(t) for t in self.__iter__()])
        return (
            f"{self.__class__.__name__}("
            f"id={self._dataclass.get_id()}, {subobjs}"
            f")"
        )

    def __eq__(self, other: object) -> bool:
        return self._dataclass == other._dataclass


class RootReport(BaseReport['ProjectReport', RootDataclass]):
    pass


class ProjectReport(BaseReport['ClientReport', ProjectDataclass]):

    def to_clients_with_best_scores_df(self) -> pd.DataFrame:
        """
        Get a dataframe with the info of each last round for each client.
        """
        df_data = []
        for client_report in self:
            client_report: ClientReport = client_report
            last_round_report: RoundReport = client_report[-1]

            last_best_result_summary: Tuple[float, ScopedLocationAttrInt, TrialStatus,
                                            FlatDict] = last_round_report.best_result_summary(use_wildcards=True)
            main_metric_name = last_round_report.main_metric_name
            df_data.append({
                ClientReport.CLIENT_ID_COLUMN_NAME: client_report.get_id(),
                f"most recent {RoundReport.ROUND_ID_COLUMN_NAME}": last_round_report.get_id(),
                'n trials for most recent round': len(last_round_report),
                'main metric name': main_metric_name,
                'best score for main metric': last_best_result_summary[0],
                'best trial status': last_best_result_summary[2].value,
                'best hyperparameters': pprint.pformat(last_best_result_summary[3])
            })

        return pd.DataFrame(df_data)


class ClientReport(BaseReport['RoundReport', ClientDataclass]):
    CLIENT_ID_COLUMN_NAME = dataclass_2_id_attr[ClientDataclass]

    def to_rounds_with_best_scores_df(self) -> pd.DataFrame:
        """
        Get a dataframe with the best scores for each round as well as the metric used for that round.
        """
        df_data = []
        for round_report in self:
            round_report: RoundReport = round_report
            best_result_summary: Tuple[float, ScopedLocationAttrInt,
                                       TrialStatus, FlatDict] = round_report.best_result_summary(use_wildcards=True)
            main_metric_name = round_report.main_metric_name
            df_data.append({
                RoundReport.ROUND_ID_COLUMN_NAME: round_report.get_id(),
                'n trials': len(round_report),
                'main metric name': main_metric_name,
                'best score for main metric': best_result_summary[0],
                'best trial status': best_result_summary[2].value,
                'best hyperparameters': pprint.pformat(best_result_summary[3])
            })

        return pd.DataFrame(df_data)


class RoundReport(BaseReport['TrialReport', RoundDataclass]):
    ROUND_ID_COLUMN_NAME = dataclass_2_id_attr[RoundDataclass]
    SUMMARY_STATUS_COLUMNS_NAME = 'status'

    @property
    def main_metric_name(self) -> str:
        return self._dataclass.main_metric_name

    def get_best_trial(self, metric_name: str = None) -> Optional['TrialReport']:
        """
        Return trial report with best score from all trials, provided that this trial has a score and was successful.
        """
        metric_name = metric_name or self.main_metric_name
        best_trial_id = self.get_best_trial_id(metric_name)
        if best_trial_id is None:
            return None
        return self[best_trial_id]

    def get_best_trial_id(self, metric_name: str = None) -> Optional[ScopedLocationAttr]:
        """
        Get best trial id from all trials. Will return None if there are no successful trial with such score.
        """
        metric_name = metric_name or self.main_metric_name
        best_score, best_trial_id = None, None
        _is_higher_score_better = self.is_higher_score_better(metric_name)

        for trial_report in self:
            trial_report: TrialReport = trial_report

            trial_score = trial_report.get_avg_validation_score(metric_name)

            _has_better_score = best_score is None or (
                trial_score is not None and trial_report.is_success() and (
                    _is_higher_score_better == (trial_score > best_score)
                )
            )

            if _has_better_score:
                best_score = trial_score
                best_trial_id = trial_report.get_id()

        return best_trial_id

    def get_best_hyperparams(self, metric_name: str = None) -> HyperparameterSamples:
        """
        Get best hyperparams from all trials.
        """
        if metric_name is None:
            metric_name = self.main_metric_name
        best_trial_report = self.get_best_trial(metric_name)
        if best_trial_report is None:
            return HyperparameterSamples()
        return best_trial_report.get_hyperparams()

    def is_higher_score_better(self, metric_name: str = None) -> bool:
        """
        Return true if higher score is better. If metric_name is None, the optimizer's
        metric is taken.
        """
        metric_name = metric_name or self.main_metric_name
        if metric_name is None:
            raise ValueError("metric name and main metric name are not defined.")

        if len(self) == 0:
            return ValueError("No trial found, cannot determine if higher score is better.")

        all_metric_names: List[List[str]] = [t.get_metric_names() for t in self]
        if metric_name not in sum(all_metric_names, []):
            raise ValueError(
                f"No trial found with the metric {metric_name}. "
                f"Trials for round {self.get_id()} have the "
                f"respective following metrics: {all_metric_names}")

        for last_with_score in reversed(range(len(self))):
            if metric_name in all_metric_names[last_with_score]:
                break

        last_trial_with_score = self[last_with_score]
        return last_trial_with_score.is_higher_score_better(metric_name)

    def get_n_val_splits(self):
        """
        Finds the number of validation splits on record in this round's first trial.
        """
        if len(self) > 0:
            return len(self[0])
        return 0

    def get_metric_names(self) -> List[str]:
        """
        Get the name of all metrics on record.
        """
        _metrics = []
        for i in self:
            i: TrialReport = i
            _ms = i.get_metric_names()
            for m in _ms:
                if m not in _metrics:
                    _metrics.append(m)
        if self.main_metric_name in _metrics:
            _metrics.remove(self.main_metric_name)
            _metrics.insert(0, self.main_metric_name)
        return _metrics

    def best_result_summary(self, metric_name: str = None, use_wildcards: bool = False) -> Tuple[float, ScopedLocationAttrInt, TrialStatus, FlatDict]:
        """
        Return the best result summary for the given metric, as the `[score, trial_number, hyperparams_flat_dict]`.
        """
        metric_name = metric_name or self.main_metric_name
        return self.summary(metric_name, use_wildcards)[0]

    def summary(
        self, metric_name: str = None, use_wildcards: bool = False
    ) -> List[Tuple[float, ScopedLocationAttrInt, FlatDict]]:
        """
        Get a summary of the round. Best score is first.
        Values in the returned triplet tuples are: (score, trial_number, hyperparams),
        sorted by score such that the best score is first.
        """
        metric_name = metric_name or self.main_metric_name
        results: List[float, ScopedLocationAttrInt, TrialStatus, FlatDict] = list()
        results_not_success: List[float, ScopedLocationAttrInt, TrialStatus, FlatDict] = list()

        for trial in self:
            trial: TrialReport = trial
            score = trial.get_avg_validation_score(metric_name)
            trial_number = trial._dataclass.trial_number
            hp = trial.get_hyperparams().to_flat_dict(use_wildcards=use_wildcards)
            status = trial.get_status()

            _resultslist = results
            if not status == TrialStatus.SUCCESS:
                _resultslist = results_not_success
            _resultslist.append((score, trial_number, status, hp))

        is_reverse: bool = self.is_higher_score_better(metric_name)
        results = list(sorted(results, reverse=is_reverse))
        results_not_success = list(sorted(results_not_success, reverse=is_reverse))
        return results + results_not_success

    def get_all_hyperparams(self, as_flat: bool = True, use_wildcards: bool = False) -> List[FlatDict]:
        """
        Get the list of hyperparams for all trials.
        """
        if use_wildcards and not as_flat:
            raise ValueError("Cannot use wildcards with non-flat hyperparams.")

        hyperparams: List[FlatDict] = list()
        for trial in self:
            hps = trial.get_hyperparams()
            if as_flat:
                hps = hps.to_flat_dict(use_wildcards=use_wildcards)
            hyperparams.append(hps)

        return hyperparams

    def list_hyperparameters_wildcards(self, discard_singles=False) -> List[str]:
        """
        Returns a list of all the hyperparameters wildcards used in the round.
        Discarding singles would prune out the hyperparameters with values that never vary.
        """
        trial_reports: List[TrialReport] = list(self)
        trials_hps = [t.get_hyperparams().to_wildcards() for t in trial_reports]
        trials_hps_keys = set()
        trials_hps_values = defaultdict(set)
        for hps in trials_hps:
            for hp, hp_val in hps.items():
                trials_hps_keys.add(hp)
                trials_hps_values[hp].add(hp_val)
        if discard_singles:
            trials_hps_keys_step_2 = [k for k, v in trials_hps_values.items() if len(v) > 1]
            return trials_hps_keys_step_2
        return list(trials_hps_keys)

    def list_successful_avg_validation_scores(self) -> List[float]:
        """
        Returns a list of all the average validation scores on record, only for succesful trials.
        """
        return [t.get_avg_validation_score(self.main_metric_name) for t in self.successful_trials]

    @property
    def successful_trials(self) -> List['TrialReport']:
        """
        Returns a list of all the succesful trials on record.
        """
        return [t for t in self if t.is_success()]

    def to_round_scatterplot_df(self, metric_name: str = None, wildcards_to_keep: List[str] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Returns a dataframe with trial ids, the selected metric, and the wildcarded hyperparameters to keep.
        """
        metric_name = metric_name or self.main_metric_name
        summary: List[float, ScopedLocationAttrInt, FlatDict] = self.summary(metric_name, use_wildcards=True)

        splom_df = pd.DataFrame([
            {**{TrialReport.TRIAL_ID_COLUMN_NAME: trial_id}, **{self.SUMMARY_STATUS_COLUMNS_NAME: status.value}, **{metric_name: score},
                **_filter_df_hps(hyperparams, wildcards_to_keep)}
            for score, trial_id, status, hyperparams in summary
        ])
        splom_df.set_index(TrialReport.TRIAL_ID_COLUMN_NAME)

        return splom_df

    def to_scores_over_time_df(self, metric_name: str = None, wildcards_to_keep: List[str] = None) -> pd.DataFrame:
        """
        Returns a dataframe with trial ids, epochs, the selected metric, and the wildcarded hyperparameters to keep.
        """
        metric_name = metric_name or self.main_metric_name
        _df_list: List[Dict(str, Any)] = []

        trials: List[TrialReport] = list(self)
        for trial in trials:
            scores_over_time: List[float] = trial.get_avg_validation_score(metric_name, over_time=True)
            if scores_over_time is None:
                continue
            hps = _filter_df_hps(trial.get_hyperparams().to_wildcards(), wildcards_to_keep)
            for epoch, s in enumerate(scores_over_time):
                _df_list.append({
                    TrialReport.TRIAL_ID_COLUMN_NAME: trial._dataclass.get_id(),
                    MetricResultsReport.EPOCH_COLUMN_NAME: epoch,
                    metric_name: s,
                    **hps,
                })

        df = pd.DataFrame(_df_list)
        df.set_index(TrialReport.TRIAL_ID_COLUMN_NAME)

        return df


class TrialReport(BaseReport['TrialSplitReport', TrialDataclass]):
    TRIAL_ID_COLUMN_NAME = dataclass_2_id_attr[TrialDataclass]

    def get_metric_names(self) -> List[str]:
        """
        Get the name of all metrics on record.
        """
        _metrics = []
        for i in self:
            i: TrialSplitReport = i
            _ms = i.get_metric_names()
            for m in _ms:
                if m not in _metrics:
                    _metrics.append(m)
        return _metrics

    def get_hyperparams(self) -> RecursiveDict:
        """
        Get the hyperparameters of the trial.
        """
        return self._dataclass.hyperparams

    def is_success(self):
        """
        Checks if the trial is successful from its dataclass record.
        """
        return self._dataclass.status == TrialStatus.SUCCESS

    def get_status(self) -> TrialStatus:
        """
        Get the status of the trial.
        """
        return self._dataclass.status

    def are_all_splits_successful(self) -> bool:
        """
        Return true if all splits are successful.
        """
        return all(i.is_success() for i in self)

    def are_all_splits_failures(self) -> bool:
        """
        Return true if all splits are failed.
        """
        return all(not i.is_success() for i in self)

    def get_avg_validation_score(self, metric_name: str, over_time=False) -> Optional[Union[float, List[float]]]:
        """
        Returns the average score for all validation splits's
        best validation score for the specified scoring metric.

        : param metric_name: The name of the metric to use.
        : param over_time: If true, return all the avg scores over time instead of the best avg score.
        : return: validation score
        """
        if self.is_success():
            scores = [
                (
                    val_split[metric_name].get_valid_scores()  # List[float]
                    if over_time is True else
                    val_split[metric_name].get_best_validation_score()  # best float
                )
                for val_split in self
                if val_split.is_success() and metric_name in val_split.get_metric_names()
            ]
            scores = [s for s in scores if s is not None]
            return np.mean(scores, axis=0) if len(scores) > 0 else None

    def get_avg_n_epoch_to_best_validation_score(self, metric_name: str) -> Optional[float]:
        if metric_name not in self._dataclass.get_sublocation_keys():
            return None

        n_epochs = [
            val_split[metric_name].get_n_epochs_to_best_validation_score()
            for val_split in self
            if val_split.is_success() and metric_name in val_split.get_metric_names()
        ]

        n_epochs = sum(n_epochs) / len(n_epochs) if len(n_epochs) > 0 else None
        return n_epochs

    def is_higher_score_better(self, metric_name: str) -> bool:
        return self[-1].is_higher_score_better(metric_name)

    def to_scores_over_time_df(self, metric_name: str) -> pd.DataFrame:
        """
        Returns a dataframe with the trial id, the epoch, and the selected metric.
        """
        _df_list: List[Dict(str, Any)] = []
        for trial_split in self:
            trial_split: TrialSplitReport = trial_split

            _split_df_list = trial_split._get_scores_over_time_train_val_df_list(metric_name)

            _df_list.extend(_split_df_list)
        df = pd.DataFrame(_df_list)
        df.set_index(TrialSplitReport.TRIAL_SPLIT_ID_COLUMN_NAME)
        return df


class TrialSplitReport(BaseReport['MetricResultsReport', TrialSplitDataclass]):
    TRIAL_SPLIT_ID_COLUMN_NAME = dataclass_2_id_attr[TrialSplitDataclass]

    def get_hyperparams(self) -> RecursiveDict:
        """
        Get the hyperparameters of the trial.
        """
        return self._dataclass.hyperparams

    def get_metric_names(self) -> List[str]:
        """
        List metric names that are the subdataclass' keys.
        """
        return list(self._dataclass.metric_results.keys())

    def is_success(self):
        """
        Set trial status to success.
        """
        return self._dataclass.status == TrialStatus.SUCCESS

    def is_higher_score_better(self, metric_name: str) -> bool:
        return self[metric_name].is_higher_score_better()

    def to_scores_over_time_df(self, metric_name: str) -> pd.DataFrame:
        """
        Returns a dataframe with the trial id, the epoch, and the selected metric.
        """
        _df_list = self._get_scores_over_time_train_val_df_list(metric_name)
        df = pd.DataFrame(_df_list)
        df.set_index(TrialSplitReport.TRIAL_SPLIT_ID_COLUMN_NAME)
        return df

    def _get_scores_over_time_train_val_df_list(self, metric_name: str) -> List[Dict[str, Any]]:
        _df_list: List[Dict(str, Any)] = []

        train_scores_over_time: List[float] = self[metric_name].get_train_scores()
        if train_scores_over_time is None:
            return []
        for epoch, s in enumerate(train_scores_over_time):
            _df_list.append({
                TrialSplitReport.TRIAL_SPLIT_ID_COLUMN_NAME: self.get_id(),
                MetricResultsReport.EPOCH_COLUMN_NAME: epoch,
                MetricResultsReport.TRAIN_VAL_COLUMN_NAME: 'train',
                metric_name: s,
            })

        valid_scores_over_time: List[float] = self[metric_name].get_valid_scores()
        if valid_scores_over_time is None:
            return _df_list
        for epoch, s in enumerate(valid_scores_over_time):
            _df_list.append({
                TrialSplitReport.TRIAL_SPLIT_ID_COLUMN_NAME: self.get_id(),
                MetricResultsReport.EPOCH_COLUMN_NAME: epoch,
                MetricResultsReport.TRAIN_VAL_COLUMN_NAME: 'validation',
                metric_name: s,
            })

        return _df_list


class MetricResultsReport(BaseReport[None, MetricResultsDataclass]):
    METRIC_COLUMN_NAME = "metric"
    EPOCH_COLUMN_NAME = "epoch"
    TRAIN_VAL_COLUMN_NAME = "phase"

    @property
    def metric_name(self) -> str:
        return self._dataclass.metric_name

    def get_train_scores(self) -> List[float]:
        """
        Return the train scores' values.
        """
        return self._dataclass.train_values

    def get_valid_scores(self) -> List[float]:
        """
        Return the validation scores for the given scoring metric.
        """
        return self._dataclass.validation_values

    def get_final_validation_score(self) -> float:
        """
        Return the latest validation score for the given scoring metric.
        """
        return self.get_valid_scores()[-1]

    def get_best_validation_score(self) -> Optional[float]:
        """
        Return the best validation score for the given scoring metric.
        """
        scores = self.get_valid_scores()
        if len(scores) == 0:
            return None

        if self.is_higher_score_better():
            f = np.max
        else:
            f = np.min

        return f(scores)

    def get_n_epochs_to_best_validation_score(self) -> Optional[int]:
        """
        Return the number of epochs
        """
        scores = self.get_valid_scores()
        if len(scores) == 0:
            return None

        if self.is_higher_score_better():
            f = np.argmax
        else:
            f = np.argmin

        return f(scores)

    def is_higher_score_better(self) -> bool:
        """
        Return True if higher scores are better for the main metric.
        """
        return self._dataclass.higher_score_is_better

    def is_new_best_score(self) -> bool:
        """
        Return True if the latest validation score is the new best score.
        """
        if self.get_best_validation_score() in self.get_valid_scores()[:-1]:
            return False
        return True


report_2_subreport: typing.OrderedDict[Type[BaseReport], Type[BaseReport]] = OrderedDict([
    (RootReport, ProjectReport),
    (ProjectReport, ClientReport),
    (ClientReport, RoundReport),
    (RoundReport, TrialReport),
    (TrialReport, TrialSplitReport),
    (TrialSplitReport, MetricResultsReport),
    (MetricResultsReport, None),
])

report_2_dataclass: typing.OrderedDict[Type[BaseReport], Type[BaseDataclass]] = OrderedDict([
    (RootReport, RootDataclass),
    (ProjectReport, ProjectDataclass),
    (ClientReport, ClientDataclass),
    (RoundReport, RoundDataclass),
    (TrialReport, TrialDataclass),
    (TrialSplitReport, TrialSplitDataclass),
    (MetricResultsReport, MetricResultsDataclass),
])
dataclass_2_report: typing.OrderedDict[Type[BaseDataclass], Type[BaseReport]] = {
    dc: rep
    for rep, dc in report_2_dataclass.items()
}



"""
Neuraxle's training callbacks classes.
=========================================
Training callback classes.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

from abc import ABC, abstractmethod
from typing import Callable, List, Optional

from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import ARG_Y_EXPECTED, ARG_Y_PREDICTD, EvalEOTDACT
from neuraxle.logging.warnings import warn_deprecated_class
from neuraxle.metaopt.data.aggregates import MetricResults, TrialSplit

OPTIONAL_ARG_CONTEXT = Optional[CX]
RETURNS_SCORE = float


class BaseCallback(ABC):
    """
    Base class for a training callback.
    Callbacks are called after each epoch inside the fit function of the :class:`~neuraxle.metaopt.automl.Trainer`.

    .. seealso::
        :class:`MetaCallback`,
        :class:`~neuraxle.metaopt.auto_ml.AutoML`,
        :class:`~neuraxle.metaopt.auto_ml.Trainer`,
    """

    @abstractmethod
    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        pass


class EarlyStoppingCallback(BaseCallback):
    """
    Perform early stopping when there is multiple epochs in a row that didn't improve the performance of the model.
    """

    def __init__(self, max_epochs_without_improvement, metric_name=None):
        """

        :param max_epochs_without_improvement: The number of step without improvement on the validation score before an early stopping is triggered.
        :param metric_name: The name of the metric on which we want to condition the early stopping. If None, the main metric will be used.
        """
        self.n_epochs_without_improvement = max_epochs_without_improvement
        self.metric_name = None

    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        metric_name = self.metric_name or trial_split.main_metric_name
        unmanaged_metric_results: MetricResults = trial_split.metric_result(metric_name)
        validation_scores = unmanaged_metric_results.get_valid_scores()

        if len(validation_scores) > self.n_epochs_without_improvement:
            higher_score_is_better = unmanaged_metric_results.is_higher_score_better()
            if (higher_score_is_better) and \
                    all(validation_scores[-self.n_epochs_without_improvement] >= v for v in
                        validation_scores[-self.n_epochs_without_improvement:]):
                return True
            elif (not higher_score_is_better) and \
                    all(validation_scores[-self.n_epochs_without_improvement] <= v for v in
                        validation_scores[-self.n_epochs_without_improvement:]):
                return True
        return False


class MetaCallback(BaseCallback):
    """
    Meta callback wraps another callback.
    It can be useful to test conditions before executing certain callbacks.

    .. seealso::
        :class:`BaseCallback`,
        :class:`IfBestScore`,
        :class:`IfLastStep`,
        :class:`StepSaverCallback`,
    """

    def __init__(self, wrapped_callback: BaseCallback):
        self.wrapped_callback = wrapped_callback

    @abstractmethod
    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        return self.wrapped_callback.call(
            trial_split,
            dact_train,
            dact_valid,
            is_finished_and_fitted
        )


class IfBestScore(MetaCallback):
    """
    Meta callback that only execute when the trial is a new best score.

    .. seealso::
        :class:`MetaCallback`,
    """

    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        if trial_split.metric_result().is_new_best_score():
            return self.wrapped_callback.call(
                trial_split,
                dact_train,
                dact_valid,
                is_finished_and_fitted
            )
        return False


class IfLastStep(MetaCallback):
    """
    Meta callback that only execute when the training is finished or fitted, or when it is the last epoch.

    .. seealso::
        :class:`MetaCallback`,
    """

    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        if trial_split.epoch == trial_split.n_epochs or is_finished_and_fitted:
            self.wrapped_callback.call(
                trial_split,
                dact_train,
                dact_valid,
                is_finished_and_fitted
            )
            return True
        return False


class StepSaverCallback(BaseCallback):
    """
    Callback that saves the trial model.

    .. seealso::
        :class:`BaseCallback`,
        :class:`~neuraxle.metaopt.auto_ml.AutoML`,
        :class:`~neuraxle.metaopt.auto_ml.Trainer`,
    """

    def __init__(self, label):
        BaseCallback.__init__(self)
        self.label = label

    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        # TODO: maybe the trial split could contain the model again. This would imply that the trainer would let the trial split train itself. This is a bit weird, still.
        trial_split.save_model(self.label)
        return False


class BestModelCheckpoint(IfBestScore):
    """
    Saves the pipeline model in a folder named "best" when the a new best validation score is reached.
    It is important to note that when refit=True, an AutoML loop will overwrite the best model after refitting.
    """

    def __init__(self):
        IfBestScore.__init__(self, StepSaverCallback('best'))


class CallbackList(BaseCallback):
    """
    Callback list that be executed.

    .. seealso::
        :class:`~neuraxle.metaopt.auto_ml.AutoML`,
        :class:`~neuraxle.metaopt.auto_ml.Trainer`,
        :class:`BaseCallback`,
    """

    def __init__(self, callbacks: List[BaseCallback]):
        self.callbacks: List[BaseCallback] = callbacks

    def __getitem__(self, item):
        return self.callbacks[item]

    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        for callback in self.callbacks:
            try:
                if callback.call(
                    trial_split=trial_split,
                    dact_train=dact_train,
                    dact_valid=dact_valid,
                    is_finished_and_fitted=is_finished_and_fitted
                ):
                    is_finished_and_fitted = True
            except Exception as e:
                trial_split.context.flow.log_error(e)
                raise e

        return is_finished_and_fitted

    def append(self, callback: BaseCallback) -> 'CallbackList':
        if callback is not None:
            self.callbacks.append(callback)
        return self


class MetricCallback(BaseCallback):
    """
    Callback that calculates metric results.
    Adds the results into the trial repository.

    .. seealso::
        :class:`BaseCallback`,
        :class:`MetaCallback`,
        :class:`EarlyStoppingCallback`,
        :class:`IfBestScore`,
        :class:`IfLastStep`,
        :class:`StepSaverCallback`,
        :class:`CallbackList`,
    """

    def __init__(
        self,
        name: str,
        metric_function: Callable[[ARG_Y_EXPECTED, ARG_Y_PREDICTD, OPTIONAL_ARG_CONTEXT], RETURNS_SCORE],
        higher_score_is_better: bool,
        log_metrics=True,
        pass_context_to_metric_function: bool = False
    ):
        # TODO: make a dictionnary somewhere to store predefined metrics?
        self.name: str = name
        self.metric_function: Callable[
            [ARG_Y_EXPECTED, ARG_Y_PREDICTD, OPTIONAL_ARG_CONTEXT], RETURNS_SCORE] = metric_function
        self.higher_score_is_better = higher_score_is_better
        self.log_metrics = log_metrics
        self.pass_context_to_metric_function = pass_context_to_metric_function

    def call(
        self,
        trial_split: TrialSplit,
        dact_train: EvalEOTDACT,
        dact_valid: EvalEOTDACT,
        is_finished_and_fitted: bool = False
    ) -> bool:
        f = self.metric_function

        if self.pass_context_to_metric_function:
            train_score: float = f(dact_train.eo, dact_train.di, trial_split.validation_context())
            if dact_valid is not None:
                valid_score: float = f(dact_valid.eo, dact_valid.di, trial_split.validation_context())
        else:
            train_score: float = f(dact_train.eo, dact_train.di)
            if dact_valid is not None:
                valid_score: float = f(dact_valid.eo, dact_valid.di)

        with trial_split.managed_metric(self.name, self.higher_score_is_better) as metric:
            metric: MetricResults = metric  # just a typing comment here for convenience.

            metric.add_train_result(train_score)
            if dact_valid is not None:
                metric.add_valid_result(valid_score)

        return is_finished_and_fitted


class ScoringCallback(MetricCallback):
    """
    Metric Callback that calculates metric results for the main scoring metric.
    Adds the results into the trial repository.

    .. seealso::
        :class:`MetricCallback`,
    """

    def __init__(
        self,
        metric_function: Callable[[ARG_Y_EXPECTED, ARG_Y_PREDICTD, OPTIONAL_ARG_CONTEXT], RETURNS_SCORE],
        name='main',
        higher_score_is_better: bool = True,
        log_metrics: bool = True,
        pass_context_to_metric_function: bool = False
    ):
        MetricCallback.__init__(
            self,
            name=name,
            metric_function=metric_function,
            higher_score_is_better=higher_score_is_better,
            log_metrics=log_metrics,
            pass_context_to_metric_function=pass_context_to_metric_function
        )
        warn_deprecated_class(self, MetricCallback)



"""
Neuraxle's AutoML Context Management
====================================================

The :class:`~neuraxle.base.ExecutionContext`, is a special object
that needs to be adapted in AutoML loops.


.. seealso::
    :class:`~neuraxle.metaopt.hyperopt.tpe.TreeParzenEstimator`,


..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""

import logging

from neuraxle.base import ExecutionContext as CX, ExecutionPhase
from neuraxle.logging.logging import NeuraxleLogger
from neuraxle.metaopt.data.vanilla import (BaseDataclass, ScopedLocation,
                                           SubDataclassT)
from neuraxle.metaopt.repositories.repo import (HyperparamsRepository, SynchronizedHyperparamsRepositoryWrapper,
                                                VanillaHyperparamsRepository)


class AutoMLContext(CX):

    @property
    def logger(self) -> NeuraxleLogger:
        # self.add_scoped_logger_file_handler()  # TODO: this is perhaps why logs are duplicated.
        return CX.logger.fget(self)

    @property
    def logger_at_scoped_loc(self) -> NeuraxleLogger:
        return logging.getLogger(self.get_identifier(include_step_names=False))

    def add_scoped_logger_file_handler(self) -> 'AutoMLContext':
        """
        Add a file handler to the logger at the current scoped location to capture logs
        at this scope and below this scope.
        """
        self.repo.add_logging_handler(self.logger_at_scoped_loc, self.loc)
        return self

    def free_scoped_logger_file_handler(self) -> 'AutoMLContext':
        """
        Remove file handlers from logger to free file lock (especially on Windows).
        """
        self.logger_at_scoped_loc.without_file_handler()
        return self

    def read_scoped_log(self) -> str:
        """
        Read the scoped logger file.
        """
        # TODO: with self.lock:
        return self.repo.get_log_from_logging_handler(self.logger, self.loc)

    def _copy(self, copy_func: str = '_copy'):
        copy_kwargs = self._get_copy_kwargs(copy_func)
        return AutoMLContext(**copy_kwargs)

    def _get_copy_kwargs(self, copy_func: str):
        return CX._get_copy_kwargs(self, copy_func)

    def new_trial(self) -> 'CX':
        """
        Set the context's execution phase to train.
        """
        new_self = self._copy(copy_func='_copy_trial')
        new_self.set_execution_phase(ExecutionPhase.PRETRAIN)
        return new_self

    def new_trial_split(self) -> 'CX':
        """
        Set the context's execution phase to train.
        """
        new_self = self._copy(copy_func='_copy_trial_split')
        new_self.set_execution_phase(ExecutionPhase.PRETRAIN)
        return new_self

    def train(self) -> 'CX':
        """
        Set the context's execution phase to train.
        """
        new_self = self._copy(copy_func='_copy_train')
        new_self.set_execution_phase(ExecutionPhase.TRAIN)
        return new_self

    def validation(self) -> 'CX':
        """
        Set the context's execution phase to validation.
        """
        new_self = self._copy(copy_func='_copy_validation')
        new_self.set_execution_phase(ExecutionPhase.VALIDATION)
        return new_self

    @staticmethod
    def from_context(
        context: CX = None,
        repo: HyperparamsRepository = None,
        loc: ScopedLocation = None,
        disable_repo_lock: bool = False,
    ) -> 'AutoMLContext':
        """
        Create a new AutoMLContext from an ExecutionContext.

        :param context: ExecutionContext
        """
        new_context: AutoMLContext = AutoMLContext._copy(
            context if context is not None else AutoMLContext()
        )
        if not new_context.has_service(HyperparamsRepository):
            new_context.register_service(
                HyperparamsRepository,
                (repo or VanillaHyperparamsRepository(new_context.get_path())).with_lock()
            )
        if not new_context.has_service(ScopedLocation):
            new_context.register_service(
                ScopedLocation,
                loc or ScopedLocation()
            )
        if not disable_repo_lock:
            new_context = new_context.synchroneous()
        return new_context

    @property
    def loc(self) -> ScopedLocation:
        return self.get_service(ScopedLocation)

    @property
    def repo(self) -> SynchronizedHyperparamsRepositoryWrapper:
        repo: HyperparamsRepository = self.get_service(HyperparamsRepository)
        return repo.with_lock()

    def push_attr(self, subdataclass: SubDataclassT) -> 'AutoMLContext':
        """
        Push a new attribute into the ScopedLocation.

        :param name: attribute name
        :param value: attribute value
        :return: an AutoMLContext copy with the new loc attribute.
        """
        if not isinstance(subdataclass, BaseDataclass) and isinstance(subdataclass, (str, int)):
            return self.with_loc(self.loc.with_id(subdataclass))  # ID instead.
        return self.with_loc(self.loc.with_dc(subdataclass))

    def pop_attr(self) -> 'AutoMLContext':
        """
        Pop an attribute from the ScopedLocation.

        :return: an AutoMLContext copy with the new popped loc attribute.
        """
        return self.with_loc(self.loc.popped())

    def with_loc(self, loc: ScopedLocation) -> 'AutoMLContext':
        """
        Replace the ScopedLocation by the one provided.

        :param loc: ScopedLocation
        :return: an AutoMLContext copy with the new loc attribute.
        """
        new_self: AutoMLContext = self._copy()
        new_self.register_service(ScopedLocation, loc)
        return new_self

    def load_dc(self, deep=True) -> BaseDataclass:
        """
        Load the current dc from the repo.
        """
        with self.repo.lock:
            return self.repo.load(self.loc, deep)



"""
Neuraxle's Logging module
====================================
This module contains the Logging class, which is used to log information about the execution of a pipeline.
It is used by the classes inheriting from BaseStep to log information.
It is also modified by the AutoML class and its Trainer and various Trial repositories-related classes
to log info in various folders.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import logging
import re
import sys
from threading import Thread
from io import StringIO
from multiprocessing import Queue
from typing import IO, Dict, List, Optional

LOGGER_FORMAT = "⠀[%(asctime)s][%(levelname)-8s][%(name)-8s][%(module)-1s.py:%(lineno)-1d][%(processName)s:%(threadName)s]: %(message)s"
LOGGER_FORMAT_PREFIX_REPLACE_REGEXPR = r"⠀(\[.*?\]): ", r""
LOGGER_FORMAT_PREFIX_SEP_L = "["
LOGGER_FORMAT_PREFIX_SEP_R = "]"
LOGGING_DATETIME_STR_FORMAT = '%Y-%m-%d_%H:%M:%S.%f'

FORMATTER = logging.Formatter(fmt=LOGGER_FORMAT, datefmt=LOGGING_DATETIME_STR_FORMAT)

if sys.version_info.major <= 3 and sys.version_info.minor <= 7:
    logging.basicConfig(format=LOGGER_FORMAT, datefmt=LOGGING_DATETIME_STR_FORMAT, level=logging.INFO)
else:
    logging.basicConfig(format=LOGGER_FORMAT, datefmt=LOGGING_DATETIME_STR_FORMAT, level=logging.INFO, force=True)

NEURAXLE_LOGGER_NAME = "neuraxle"


LOGGER_STRING_IO: Dict[str, StringIO] = {}
LOGGER_FILE_HANDLERS: Dict[str, logging.FileHandler] = {}


class _FilterSTDErr(logging.Filter):
    def filter(self, record):
        return record.levelno < logging.WARN


class NeuraxleLogger(logging.Logger):

    @staticmethod
    def root() -> 'NeuraxleLogger':
        return NeuraxleLogger(NEURAXLE_LOGGER_NAME)

    @staticmethod
    def from_identifier(identifier: str) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        logger: NeuraxleLogger = logging.getLogger(identifier)

        logger.with_string_io(identifier)
        return logger

    def with_string_io(self, identifier: str) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        if identifier not in LOGGER_STRING_IO:
            string_io = StringIO()
            LOGGER_STRING_IO[identifier] = string_io
            self._add_stream_handler("string_io_handler", string_io)
        return self

    def with_file_handler(self, file_path: str) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        self.without_file_handler()

        file_handler = logging.FileHandler(file_path)
        file_handler.setLevel('DEBUG')
        self._add_partial_handler(f"file_handler:{file_path}", file_handler, level=logging.DEBUG)

        LOGGER_FILE_HANDLERS[self.name] = file_handler
        return self

    def without_file_handler(self) -> 'NeuraxleLogger':
        """
        Returns a logger from an identifier.
        :param identifier: The identifier of the logger
        :return: The logger
        """
        if self.name in LOGGER_FILE_HANDLERS:
            LOGGER_FILE_HANDLERS[self.name].close()
            try:
                self.removeHandler(LOGGER_FILE_HANDLERS[self.name])
            except KeyError as ke:
                raise ke from ke  # Breakpoint here.
            del LOGGER_FILE_HANDLERS[self.name]
        return self

    def read_log_file(self) -> List[str]:
        if self.name not in LOGGER_FILE_HANDLERS:
            raise ValueError(
                f"No file handler for logger named `{self.name}`. Perhaps you forgot to call "
                f"`self.with_file_handler`, or removed the file handler?")
        with open(LOGGER_FILE_HANDLERS[self.name].baseFilename, 'r') as f:
            return f.readlines()

    def with_std_handlers(self) -> 'NeuraxleLogger':
        self._add_stream_handler("errr_handler", sys.stderr, logging.WARN)
        self._add_stream_handler("info_handler", sys.stdout, logging.DEBUG, _FilterSTDErr())
        return self

    def get_scoped_string_history(self) -> str:
        return LOGGER_STRING_IO[self.name].getvalue()

    def get_root_string_history(self) -> str:
        return LOGGER_STRING_IO[NEURAXLE_LOGGER_NAME].getvalue()

    def get_short_scoped_logs(self) -> List[str]:
        _logs = self.get_scoped_string_history()
        _logs = self.shorten_log_lines_prefixes(_logs)
        return _logs.split("\n")

    def get_short_root_logs(self) -> List[str]:
        _logs = self.get_root_string_history()
        _logs = self.shorten_log_lines_prefixes(_logs)
        return _logs.split("\n")

    def __iter__(self):
        """
        Short method to access :func:`get_short_scoped_logs` 's iterator.
        """
        return self.get_short_scoped_logs().__iter__()

    def __getitem__(self, item):
        """
        Short method to access :func:`get_short_scoped_logs` 's items.
        """
        return self.get_short_scoped_logs()[item]

    def print_root_string_history(self) -> None:
        print(f"{NEURAXLE_LOGGER_NAME} -> str:\n\n"
              f"{LOGGER_STRING_IO[NEURAXLE_LOGGER_NAME].getvalue()}")

    def _add_stream_handler(
        self,
        handler_name: str,
        stream: IO,
        level: Optional[int] = None,
        _filter: Optional[logging.Filter] = None
    ) -> 'NeuraxleLogger':
        handler = logging.StreamHandler(stream=stream)
        return self._add_partial_handler(handler_name, handler, level, _filter)

    def _add_partial_handler(
        self,
        handler_name: str,
        handler: logging.Handler,
        level: Optional[int] = None,
        _filter: Optional[logging.Filter] = None
    ) -> 'NeuraxleLogger':
        handler.setFormatter(FORMATTER)
        handler.set_name(f"{self.name}.{handler_name}")
        if level is not None:
            handler.setLevel(level)
        if _filter is not None:
            handler.addFilter(_filter)
        self.addHandler(handler)
        return self

    @staticmethod
    def shorten_log_lines_prefixes(logs):
        # This line is for retrocompability with old logs of 0.7.0.
        c = "⠀"
        logs = (c + logs.replace("\n", "\n" + c)).replace(c + c, c)
        logs = re.sub(LOGGER_FORMAT_PREFIX_REPLACE_REGEXPR[0], LOGGER_FORMAT_PREFIX_REPLACE_REGEXPR[1], logs)
        return logs.strip().rstrip(c).rstrip("\n")


logging.setLoggerClass(NeuraxleLogger)

NEURAXLE_ROOT_LOGGER: NeuraxleLogger = NeuraxleLogger.from_identifier(NEURAXLE_LOGGER_NAME)
NEURAXLE_ROOT_LOGGER.setLevel(logging.DEBUG)
NEURAXLE_ROOT_LOGGER.with_std_handlers()


class ParallelLoggingConsumerThread:
    """
    This class is used to receive logging messages sent from worker processes that
    are running in different PROCESSES (not threads).

    The logging will work when using threads. However, when using processes (multiprocessing.Process) :class:`multiprocessing.Process`
    , the present class
    is needed in the main process to receive the logging messages from the worker processes properly in the main console output.
    """

    def __init__(self, logging_queue: Queue = None):
        # type is, to me more precise: Queue[logging.LogRecord]
        self.logging_queue: Queue = logging_queue or Queue()
        self.logging_thread: Thread = None

    def start(self):
        """
        Start the thread to read the logs from the queue.
        """
        _logging_thread = Thread(
            target=self.logger_consumer_thread_func,
            args=(self.logging_queue,)
        )
        _logging_thread.daemon = True
        _logging_thread.start()
        self.logging_thread = _logging_thread

    @staticmethod
    def logger_consumer_thread_func(queue: Queue):
        """
        This function is destined to be run in a separate thread.
        Receiving a None signal in the queue means it's time to finish and to break the
        while True. This None signal is as sent from within the self. :func:`join` () method.
        The Queue normally consumes logging.LogRecord items.
        """
        while True:
            rec: logging.LogRecord = queue.get()

            if rec is None:
                break

            logger = logging.getLogger(rec.name)
            logger.handle(rec)

    def join(self, timeout: float = None):
        """
        Send the None signal for the thread to finish, and join on the thread to finish it.
        Note that once this method is called, the thread will finish and the threading.Thread
        object will be destroyed. Therefore, this method should be called only once and after
        the producers of parallel logging messages finished their job.

        Timeout is in seconds.
        """
        self.logging_queue.put(None)
        self.logging_thread.join(timeout=timeout)
        self.logging_queue.close()
        self.logging_queue.cancel_join_thread()
        self.logging_queue.join_thread()


def register_log_producer_for_main_logger_thread_to_consume(logging_queue: Queue):
    if logging_queue is not None:
        queue_handler = logging.handlers.QueueHandler(logging_queue)
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)
        root.addHandler(queue_handler)
